
"""
cgm - package for decoding cgm pages

Maybe something like this?
* parse_time
* decoders - decoder classes?
* PageDecoder - stateful decoder
"""

from binascii import hexlify

# TODO: all this stuff could be refactored into re-usable and tested
# module.
import io
from datetime import datetime
from dateutil.relativedelta import relativedelta

from decocare import lib
from decocare.records import times
from decocare.errors import DataTransferCorruptionError
from pprint import pprint

###################
#
# Time parser
# TODO: document
# TODO: tests, ideally ones showing bits flip over
###################

def parse_minutes (one):
  minute = (one & 0b111111 )
  return minute

def parse_hours (one):
  return (one & 0x1F )

def parse_day (one):
  return one & 0x1F

def parse_months (first_byte, second_byte):
  first_two_bits  = first_byte  >> 6
  second_two_bits = second_byte >> 6
  return (first_two_bits << 2) + second_two_bits
  

def parse_date (data, unmask=False, strict=False, minute_specific=False):
  """
  Some dates are formatted/stored down to the second (Sensor CalBGForPH) while
    others are stored down to the minute (CGM SensorTimestamp dates).
  """
  data = data[:]
  seconds = 0
  minutes = 0
  hours   = 0
  
  year    = times.parse_years(data[0])
  day     = parse_day(data[1])
  minutes = parse_minutes(data[2])

  hours   = parse_hours(data[3])

  month   = parse_months(data[3], data[2])

  try:
    date = datetime(year, month, day, hours, minutes, seconds)
    return date
  except ValueError, e:
    if strict:
      raise
    if unmask:
      return (year, month, day, hours, minutes, seconds)
    pass
  return None

class PagedData (object):
  """
    PagedData - context for parsing a page of cgm data.
  """

  @classmethod
  def Data (klass, data, **kwds):
    stream = io.BufferedReader(io.BytesIO(data))
    return klass(stream, **kwds)
  def __init__ (self, stream, larger=False):
    raw = bytearray(stream.read(1024))
    data, crc = raw[0:1022], raw[1022:]
    computed = lib.CRC16CCITT.compute(bytearray(data))
    self.larger = larger
    if lib.BangInt(crc) != computed:
      raise DataTransferCorruptionError("CRC does not match page data")
    
    data.reverse( )
    self.data = self.eat_nulls(data)
    self.stream = io.BufferedReader(io.BytesIO(self.data))

  def eat_nulls (self, data):
    i = 0
    while data[i] == 0x00:
      i = i+1
    return data[i:]
  def suggest (self, op):
    """
    return a partially filled in acket/opcode
     info: name, packet size, date_type, op
     some info will be added later when the record is parsed:
       GlucoseSensorData, cal factor. amount, prefix, data
    """
    # TODO: would it be possible to turn these into classes which know hwo
    # to decode time and describe themselves to PagedData?
    # that way we can write tests for each type individually and tests for
    # the thing as a whole easily.
    records = {
      0x01: dict(name='DataEnd',packet_size=0,date_type='none',op='0x01'),
      0x02: dict(name='SensorWeakSignal',packet_size=0,date_type='prevTimestamp',op='0x02'),
      0x03: dict(name='SensorCal',packet_size=1,date_type='prevTimestamp',op='0x03'),
      0x07: dict(name='Fokko-07',packet_size=1,date_type='prevTimestamp',op='0x07'),
      0x08: dict(name='SensorTimestamp',packet_size=4,date_type='minSpecific',op='0x08'),
      0x0a: dict(name='BatteryChange',packet_size=4,date_type='minSpecific',op='0x0a'),
      0x0b: dict(name='SensorStatus',packet_size=4,date_type='minSpecific',op='0x0b'),
      0x0c: dict(name='DateTimeChange',packet_size=4,date_type='secSpecific',op='0x0c'),
      0x0d: dict(name='SensorSync',packet_size=4,date_type='minSpecific',op='0x0d'),
      0x0e: dict(name='CalBGForGH',packet_size=5,date_type='minSpecific',op='0x0e'),
      0x0f: dict(name='SensorCalFactor',packet_size=6,date_type='minSpecific',op='0x0f'),
      # 0x10: dict(name='10-Something',packet_size=7,date_type='minSpecific',op='0x10'),
      0x10: dict(name='10-Something',packet_size=4,date_type='minSpecific',op='0x10'),
      0x13: dict(name='19-Something',packet_size=0,date_type='prevTimestamp',op='0x13')
    }
    if self.larger:
      # record[08][]
      pass

    if op > 0 and op < 20:
      record = records.get(op, None)
      if record is None:
        return dict(name='Could Not Decode',packet_size=0,op=op)
      else:
        return record
    else:
      record = dict(name='GlucoseSensorData',packet_size=0,date_type='prevTimestamp',op=op)
      record.update(sgv=(int(op) * 2))
      return record

  def decode (self):
    """
      XXX: buggy code

        * fails to acknowledge gaps in time
        * fails to acknowledge SensorSync
    """
    records = [ ]
    prefix_records = []
    for B in iter(lambda: self.stream.read(1), ""):
      B = bytearray(B)
      record = self.suggest(B[0])
      record['_tell'] = self.stream.tell( )
      # read packet if needed
      if not record is None and record['packet_size'] > 0:
        raw_packet = bytearray(self.stream.read(record['packet_size']))

      if record['name'] == 'DataEnd':
        prefix_records.append(record)
        continue
      
      elif record['name'] == 'GlucoseSensorData' or record['name'] == 'SensorWeakSignal' \
        or record['name'] == 'SensorCal' or record['name'] == '19-Something':
        # add to prefixed records to add to the next sensor minute timestamped record
        if record['name'] == 'SensorCal':
          record.update(raw=self.byte_to_str(raw_packet))
          if int(raw_packet[0]) == 1:
            record.update(waiting='waiting')
          else:
            record.update(waiting='meter_bg_now')
        prefix_records.append(record)

      elif record['name'] == 'SensorTimestamp' or record['name'] == 'SensorCalFactor' or record['name'] in ['10-Something' ]:
        # TODO: maybe this is like a ResetGlucose base command
        # these are sensor minute timestamped records thus create the record
        # and map prefixed elements based on the timedelta
        record.update(raw=self.byte_to_str(raw_packet))
        date, body = raw_packet[:4], raw_packet[4:]  
        date.reverse()
        date = parse_date(date)
        if date:
          record.update(date=date.isoformat())
        else:
          print "@@@", self.stream.tell( )
          pprint(dict(raw=hexlify(raw_packet)))
          pprint(dict(date=hexlify(date or bytearray( ))))
          pprint(dict(body=hexlify(body)))
          break
        prefix_records.reverse()
        mapped_glucose_records = self.map_glucose(prefix_records, start=date, delta=self.delta_ago(reverse=True))
        mapped_glucose_records.reverse()
        # And this ResetGlucose has a payload indicating calibration factor
        # Update sensor cal factor
        if record['name'] == 'SensorCalFactor': 
          factor = lib.BangInt([ body[0], body[1] ]) / 1000.0
          record.update(factor=factor) 
        records.extend(mapped_glucose_records)
        records.append(record)
        prefix_records = []


      elif record['name'] in ['SensorStatus', 'DateTimeChange', 'SensorSync', '10-Something', 'CalBGForGH', 'BatteryChange' ]:
        # independent record => parse and add to records list
        record.update(raw=self.byte_to_str(raw_packet))
        if record['name'] in ['SensorStatus', 'SensorSync', 'CalBGForGH', 'BatteryChange', 'DateTimeChange']:
          date, body = raw_packet[:4], raw_packet[4:]
          date.reverse()
          date = parse_date(date)
          if date is not None:
            record.update(date=date.isoformat())
          else:
            record.update(_date=str(raw_packet[:4]).encode('hex'))
          record.update(body=self.byte_to_str(body))
          # Update cal amount
          if record['name'] == 'DateTimeChange':
            """
            changed = body[1:5]
            changed.reverse( )
            changed = parse_date(changed)
            record.update(change=changed.isoformat( ), body=self.byte_to_str(body[5:]))
            """
          if record['name'] == 'CalBGForGH':
            amount = lib.BangInt([ (raw_packet[2] & 0b00100000) >> 5, body[0] ])
            record.update(body=self.byte_to_str(body))
            record.update(amount=amount)
        records.append(record)
      else:
        # could not decode
        records.append(record)
      # End For  
    records.reverse()
    self.records = records
    return self.records


  def byte_to_str (self, byte_array):
    # convert byte array to a string
    hex_bytes = []
    for i in range(0, len(byte_array)):
      hex_bytes.append('{0:02x}'.format(byte_array[i]))
    return '-'.join(hex_bytes)

  def map_glucose (self, values, start=None, delta=None):
    last = start
    if delta is None:
      delta = self.delta_ago()
    for x in list(values):
      if x['name'] != '19-Something':
        last = last - delta
      x.update(date=last.isoformat())
    return values
          
  def delta_ago (self, reverse=False, offset=1):
    delta = relativedelta(minutes=5*offset)
    if reverse:
      delta = relativedelta(minutes=-5*offset)
    return delta





"""
decocare - a pure python module for serial communication with insulin pumps.

This module implements the communication protocol necessary to audit
or command MM insulin pumps.  Tested with 515, and 522.

See:

.. hlist::
  * https://github.com/bewest/decoding-carelink
  * https://github.com/bewest/insulaudit
  * https://github.com/bewest/insulaudit-ssh-tools
  * https://github.com/bewest/diabetes
  * https://github.com/medevice-users/diabetes#projects


"""




#!/usr/bin/env python
# PYTHON_ARGCOMPLETE_OK

import sys
import argparse, argcomplete
import textwrap

from pprint import pprint, pformat
from binascii import hexlify
# from datetime import datetime
# from scapy.all import *
import json

from decocare import lib, history, models

from decocare.history import parse_record, HistoryPage

def get_model (spec):
  return models.known.get(spec, models.PumpModel)(spec, None)

def get_opt_parser( ):
  parser = argparse.ArgumentParser( )
  parser.add_argument('infile', nargs="+",
                      default=sys.stdin,
                      type=argparse.FileType('r'),
                      help="Find dates in this file.")

  parser.add_argument('--collate',
                      dest='collate',
                      default=False,
                      action='store_true')

  parser.add_argument('--data',
                      choices=['glucose', 'pump'],
                      default='pump')
  parser.add_argument('--model',
                      # type=get_model,
                      choices=models.known.keys( ))
  parser.add_argument('--larger',
                      dest='larger', action='store_true')
  parser.add_argument('--no-larger',
                      dest='larger', action='store_false')

  parser.add_argument('--out',
                      default=sys.stdout,
                      type=argparse.FileType('w'),
                      help="Write records here.")
  parser.set_defaults(larger=False)
  argcomplete.autocomplete(parser)
  return parser

##
# move to history.py
#


def eat_nulls(fd):
  nulls = bytearray( )
  for B in iter(lambda: bytearray(fd.read(1)), bytearray("")):
    if B[0] == 0x00:
      nulls.extend(B)
    else:
      fd.seek(fd.tell( ) - 1)
      break
  print "found %s nulls" % len(nulls)
  return nulls

def find_records(stream, opts):
  records = [ ]
  errors  = [ ]
  bolus = bytearray( )
  extra = bytearray( )
  opcode = ''

  for B in iter(lambda: bytearray(stream.read(2)), bytearray("")):

    if B == bytearray( [ 0x00, 0x00 ] ):
      print ("#### STOPPING DOUBLE NULLS @ %s," % stream.tell( )),
      nulls = eat_nulls(stream)
      print "reading more to debug %#04x" % B[0]
      print lib.hexdump(B, indent=4)
      print lib.int_dump(B, indent=11)

      extra = bytearray(stream.read(32))
      print "##### DEBUG HEX"
      print lib.hexdump(extra, indent=4)
      print "##### DEBUG DECIMAL"
      print lib.int_dump(extra, indent=11)
      # print "XXX:???:XXX", history.parse_date(bolus).isoformat( )
      break
    record = parse_record( stream, B, model=opts.model, larger=opts.larger )
    records.append(record)

  return records

import collections
Response = collections.namedtuple('Response', 'data')

def main( ):
  parser = get_opt_parser( )
  opts = parser.parse_args( )
  opts.model = get_model(opts.model)
  tw_opts = {
    'width': 50,
    'subsequent_indent': '          ',
    'initial_indent': '       ',
  }
  wrapper = textwrap.TextWrapper(**tw_opts)
  records = [ ]
  for stream in opts.infile:
    print "## START %s" % (stream.name)
    if opts.collate and opts.data == 'pump':
      page = HistoryPage(bytearray(stream.read( )), opts.model)
      records.extend(page.decode(larger=opts.larger ))
    elif opts.data == 'glucose':
      page = Response(data=bytearray(stream.read( )))
      records.extend(opts.model.iter_glucose_pages.Cursor(opts.model).find_records(page))
    else:
      records = find_records(stream, opts)
    i = 0
    for record in records:

      prefix = '#### RECORD {} {}'.format(i, str(record))
      if getattr(record, 'pformat', None):
        print record.pformat(prefix)
      else:
        # json.dumps(record, indent=2)
        print prefix
      i += 1
    print "`end %s: %s records`" % (stream.name, len(records))
    stream.close( )
  if opts.collate:
    opts.out.write(json.dumps(records, indent=2))

if __name__ == '__main__':
  import doctest
  failures, tests = doctest.testmod( )
  if failures > 0:
    print "REFUSING TO RUN DUE TO FAILED TESTS"
    sys.exit(1)
  main( )
#####
# EOF



#!/usr/bin/env python
# PYTHON_ARGCOMPLETE_OK

from decocare import commands
import io
import json
import argparse

from datetime import datetime
from dateutil import relativedelta
from dateutil.parser import parse
from dateutil.tz import gettz

from decocare import lib
from decocare.history import parse_record, HistoryPage
from decocare.helpers import cli

class LatestActivity (cli.CommandApp):
  """%(prog)s - Grab latest activity


  Query pump for latest activity.
  """
  def customize_parser (self, parser):
    parser.add_argument('--no-clock',
            dest="clock",
            action="store_false",
            default=True,
            help="Also report current time on pump."
          )
    parser.add_argument('--no-basal',
            dest="basal",
            action="store_false",
            default=True,
            help="Also report basal rates."
          )
    parser.add_argument('--no-temp',
            dest="temp",
            action="store_false",
            default=True,
            help="Also report temp basal rates."
          )
    parser.add_argument('--no-reservoir',
            dest="reservoir",
            action="store_false",
            default=True,
            help="Also report remaining insulin in reservoir."
          )
    parser.add_argument('--no-status',
            dest="status",
            action="store_false",
            default=True,
            help="Also report current suspend/bolus status."
          )
    parser.add_argument('--parser-out',
            dest="parsed_data",
            default='-',
            type=argparse.FileType('w'),
            help="Put history json in this file"
          )
    parser.add_argument('--rtc-out',
            dest="rtc_archive",
            default='-',
            type=argparse.FileType('w'),
            help="Put clock json in this file"
          )
    parser.add_argument('--reservoir-out',
            dest="reservoir_archive",
            default='-',
            type=argparse.FileType('w'),
            help="Put reservoir json in this file"
          )
    parser.add_argument('--settings-out',
            dest="settings",
            default='-',
            type=argparse.FileType('w'),
            help="Put settings json in this file"
          )
    parser.add_argument('--temp-basal-status-out',
            dest="tempbasal",
            default='-',
            type=argparse.FileType('w'),
            help="Put temp basal status json in this file"
          )
    parser.add_argument('--basals-out',
            dest="basals",
            default='-',
            type=argparse.FileType('w'),
            help="Put basal schedules json in this file"
          )
    parser.add_argument('--status-out',
            dest="status",
            default='-',
            type=argparse.FileType('w'),
            help="Put status json in this file"
          )
    parser.add_argument('--timezone',
            default=gettz( ),
            type=gettz,
            help="Timezone to use"
          )
    parser.add_argument('minutes',
            type=int,
            nargs="?",
            default=30,
            help="[default: %(default)s)]"
          )
    return parser


  def report_clock (self, args):
    self.clock = self.exec_request(self.pump, commands.ReadRTC)
    self.time = lib.parse.date(self.clock.getData( ))
    self.time = self.time.replace(tzinfo=args.timezone)
    self.timezone = args.timezone
    self.since = self.time - self.delta
    results = dict(now=self.time.isoformat( )
              , observed_at=datetime.now(args.timezone).isoformat( )
              , model=self.pump.modelNumber
              , _type='RTC')

    print "```json"
    args.rtc_archive.write(json.dumps(results, indent=2))
    print ''
    print "```"

  def report_status (self, args):
    status = self.exec_request(self.pump, commands.ReadPumpStatus)
    self.status = status.getData( )
    args.status.write(json.dumps(self.status, indent=2))

  def report_temp (self, args):
    temp = self.exec_request(self.pump, commands.ReadBasalTemp)
    self.temp = temp.getData( )
    args.tempbasal.write(json.dumps(self.temp, indent=2))
    
  def report_settings (self, args):
    settings = self.exec_request(self.pump, commands.ReadSettings)
    self.settings = settings.getData( )
    args.settings.write(json.dumps(self.settings, indent=2))

  def report_reservoir (self, args):
    reservoir = self.exec_request(self.pump, commands.ReadRemainingInsulin)
    self.reservoir = reservoir.getData( )
    args.reservoir_archive.write(json.dumps(self.reservoir, indent=2))

  def report_basal (self, args):
    profile = self.settings['selected_pattern']
    query = { 0: commands.ReadProfile_STD512
            , 1: commands.ReadProfile_A512
            , 2: commands.ReadProfile_B512
            }
    basals = self.exec_request(self.pump, query[profile])
    self.basals = basals.getData( )
    args.basals.write(json.dumps(self.basals, indent=2))

  def download_page (self, number):
    kwds = dict(page=number)
    page = self.exec_request(self.pump, commands.ReadHistoryData, args=kwds)
    return page

  def find_records (self, page, larger=None):
    decoder = HistoryPage(page, self.pump.model)
    records = decoder.decode( )
    print "SINCE", self.since.isoformat( )
    for record in records:
      print "  * found record", record['_type'], record.get('timestamp')
      print "  * should quit", record.get('timestamp') < self.since.isoformat( ), self.enough_history
      if record.get('timestamp'):
        dt = parse(record['timestamp'])
        dt = dt.replace(tzinfo=self.timezone)
        record.update(timestamp=dt.isoformat( ))
        if record['timestamp'] < self.since.isoformat( ):
          self.enough_history = True
        if record['timestamp'] >= self.since.isoformat( ):
          self.records.append(record)
    return records

  def download_history (self, args):
    i = 0
    print "find records since", self.since.isoformat( )
    self.enough_history = False
    self.records = [ ]
    while not self.enough_history:
      history = self.download_page(i)
      remainder = self.find_records(history.data)
      i = i + 1
    results = self.records
    print "```json"
    args.parsed_data.write(json.dumps(results, indent=2))
    print ''
    print "```"

  def main (self, args):
    self.delta = relativedelta.relativedelta(minutes=args.minutes)
    self.report_settings(args)
    if args.clock:
      self.report_clock(args )
    if args.status:
      self.report_status(args)
    if args.temp:
      self.report_temp(args)
    if args.basal:
      self.report_basal(args)
    if args.reservoir:
      self.report_reservoir(args)
    self.download_history(args)

if __name__ == '__main__':
  app = LatestActivity( )
  app.run(None)




import sys, os
import logging
import time
import argparse
from pprint import pformat

log = logging.getLogger( ).getChild(__name__)

from decocare import link, stick, session, commands, lib, scan

class CommandApp(object):
  def __init__(self):
    self.env = self.parse_env( )
    self.parser = self.get_parser( )
    self.autocomplete( )

  @classmethod
  def parse_env (klass):
    return {
     "serial": os.environ.get('SERIAL', ''),
     "port": os.environ.get('PORT', '')
    }

  def autocomplete (self):
    try:
      import argcomplete
      argcomplete.autocomplete(self.parser)
    except ImportError:
      # no auto completion
      pass

  def help (self):
    return self.__class__.__doc__

  def get_parser (self):
    helptext = self.help( ).split("\n")
    description = '\n'.join(helptext[0:2])
    epilog = '\n'.join(helptext[2:])
    parser = argparse.ArgumentParser(description=description, epilog=epilog)
    parser.add_argument('--serial', type=str,
                        dest='serial',
                        default=self.env.get('serial', ''),
                        help="serial number of pump [default: %(default)s]")
    parser.add_argument('--port', type=str,
                        dest='port',
                        default=(self.env.get('port') or 'scan'),
                        help="Path to device [default: %(default)s]")
    parser.add_argument('--no-op',
                        dest='dryrun',
                        action='store_true', default=False,
                        help="Dry run, don't do main function"
                        )
    parser.add_argument('--skip-prelude',
                        dest='no_prelude',
                        action='store_true', default=False,
                        help="Don't do the normal prelude."
                        )
    parser.add_argument('--no-rf-prelude',
                        dest='no_rf_prelude',
                        action='store_true', default=False,
                        help="Do the prelude, but don't query the pump."
                        )
    parser.add_argument('--skip-postlude',
                        dest='no_postlude',
                        action='store_true', default=False,
                        help="Don't do the normal postlude."
                        )
    parser.add_argument('-v', '--verbose',
                        dest='verbose',
                        action='append_const', const=1,
                        help="Verbosity"
                        )
    parser.add_argument('--rf-minutes',
                        dest='session_life',
                        type=int, default=10,
                        help="How long RF sessions should last"
                        )
    parser.add_argument('--auto-init',
                        dest='autoinit',
                        action='store_true', default=False,
                        help="Send power ctrl to initialize RF session."
                        )
    parser.add_argument('--init',
                        dest='init',
                        action='store_true', default=False,
                        help="Send power ctrl to initialize RF session."
                        )
    parser = self.customize_parser(parser)
    return parser

  def customize_parser (self, parser):
    parser.add_argument('commands',
                        nargs="+",
                        #dest='status',
                        choices=['act', 'esc', 'up', 'down', 'easy' ],
                        default='act',
                        help="buttons to press [default: %(default)s)]"
                        )
    return parser

  def configure (self):
    args = self.parser.parse_args( )
    level = None
    if args.verbose > 0:
      level = args.verbose > 1 and logging.DEBUG or logging.INFO
    logging.basicConfig(stream=sys.stdout, level=level)
    self.args = args
    return args

  def run (self, args):
    if not args:
      args = self.configure( )
    self.prelude(args)
    self.main(args)
    self.postlude(args)

  def prelude (self, args):
    if args.no_prelude:
      print "##### skipping prelude"
      return
    print "## do stuff with an insulin pump over RF"
    print "using", "`", args, "`"

    print "```"
    if args.port == 'scan' or args.port == "":
      args.port = scan.scan( )
    uart = stick.Stick(link.Link(args.port))
    print "```"
    print "```"
    uart.open( )
    print "```"
    print "```"
    pump = session.Pump(uart, args.serial)
    print "```"
    print "```"
    stats = uart.interface_stats( )
    print "```"
    print "```javascript"
    print pformat(stats)
    print "```"
    self.uart = uart
    self.pump = pump
    self.model = None
    if args.no_rf_prelude:
      print "##### skipping normal RF preludes"
      return
    print "```"
    if not args.autoinit:
      if args.init:
        pump.power_control(minutes=args.session_life)
      model = pump.read_model( )
      self.model = model
    else:
      self.autoinit(args)
    print "```"
    print '### PUMP MODEL: `%s`' % self.model

  def autoinit (self, args):
    for n in xrange(3):
      print "AUTO INIT", n
      self.sniff_model( )
      if len(self.model.getData( )) != 3:
        self.stats
        print "SENDING POWER ON", n
        self.pump.power_control(minutes=args.session_life)
      else:
        print '### PUMP MODEL: `%s`' % self.model
        break
  def sniff_model (self):
    model = self.pump.read_model( )
    print "GOT MODEL", model
    self.model = model
  def postlude (self, args):
    if args.no_postlude:
      print "##### skipping postlude"
      return
    print "### end stats"
    self.stats( )
  def stats (self):
    print "```"
    stats = self.uart.interface_stats( )
    print "```"
    print "```javascript"
    print pformat(stats)
    print "```"

  def main (self, args):
    for flow in args.commands:
      print '### ', flow
      if args.dryrun:
        print "#### dry run, no action taken"
      else:
        self.exec_request(self.pump, flow)


  def exec_request (self, pump, msg, args={},
                    dryrun=False,
                    save=False,
                    prefix='',
                    render_decoded=True,
                    render_hexdump=True):
    if dryrun:
      print "skipping query", pump, msg, args
      return False
    response = pump.query(msg, **args)
    print "response: %s" % response
    if render_hexdump:
      print "hexdump:"
      print "```python"
      print response.hexdump( )
      print "```"
    if render_decoded:
      print "#### decoded:"
      print "```python"
      print repr(response.getData( ))
      print "```"
    if save:
      response.save(prefix=prefix)
    return response




#!/usr/bin/env python
# PYTHON_ARGCOMPLETE_OK

from decocare import commands
from decocare.helpers import cli

class SetSuspendResumeApp (cli.CommandApp):
  """ %(prog)s -  query or set suspend/resume status

  Pause or resume pump.
  """
  def customize_parser (self, parser):
    parser.add_argument('commands',
                        nargs="+",
                        choices=['query', 'suspend', 'resume'],
                        default='query',
                        help="Set or query pump status [default: %(default)s)]"
                        )
    return parser

  def exec_request (self, pump, msg, **kwds):
    msg = lookup_command(msg)
    super(SetSuspendResumeApp, self).exec_request(pump, msg, **kwds)

command_map = {
    'query': commands.ReadPumpStatus
  , 'suspend': commands.PumpSuspend
  , 'resume': commands.PumpResume
}

def lookup_command (name):
  return command_map.get(name)

if __name__ == '__main__':
  app = SetSuspendResumeApp( )
  app.run(None)





import glob

class ID:
  VENDOR  = 0x0a21
  PRODUCT = 0x8001
  @classmethod
  def template (Klass, prefix):
    postfix = '*'
    usb_id = "%04x_%04x" % (ID.VENDOR, ID.PRODUCT)
    candidate = ''.join([prefix, usb_id, postfix])
    return candidate

def scan (prefix='/dev/serial/by-id/*-', template=ID.template):
  candidate = template(prefix)
  results = glob.glob(candidate)
  return (results[0:1] or ['']).pop( )

if __name__ == '__main__':
  candidate = scan( )
  print candidate





import logging
import time

import lib

class BadResponse (Exception):
  pass

"""

Implementation and decoding of lots of commands.

Each command inherits from :py:`BaseCommand`, which takes care of the
basic logic for informing the stick if we have recieved all the data
we expect to recieve.

Many commands are supported by Medtronic but not listed here.
Examples would include setting profiles and rates.
(One theory is that these commands are turned into setters with
correct arguments.)

"""

log = logging.getLogger( ).getChild(__name__)

def CRC8(data):
  return lib.CRC8.compute(data)

class BaseCommand(object):
  code    = 0x00
  descr   = "(error)"
  retries = 2
  timeout = 3
  params  = [ ]
  bytesPerRecord = 0
  maxRecords = 0
  effectTime = 0

  responded = False

  def __init__(self, code, descr, *args):
    self.code   = code
    self.descr  = descr
    self.params = [ ]

  def done(self):
    found = len(self.data or [ ])
    expect = int(self.maxRecords * self.bytesPerRecord)
    expect_size = "found[{}] expected[{}]".format(found, expect)
    log.info("%s:download:done?explain=%s" % (self, expect_size))
    return found >= expect
  def format(self):
    pass

  def respond(self, data):
    if getattr(self, 'data', None):
      self.data.extend(data)
    else:
      self.data = data
    self.getData( )
    self.responded = True

  def hexdump (self):
    return lib.hexdump(self.data)

class FieldChecker (object):
  def __init__ (self, msg, required=[]):
    self.msg = msg
    self.required = required

  def check_fields (self, data):
    for field in self.required:
      if field not in data:
        raise BadResponse( )
  def __call__ (self, data):
    self.msg.validate(data)
    self.check_fields(data)
    return True

class PumpCommand(BaseCommand):
  #serial = '665455'
  #serial = '206525'
  serial = '208850'

  params = [ ]
  bytesPerRecord = 64
  maxRecords = 1
  retries = 2
  effectTime = .500
  data = bytearray( )
  Validator = FieldChecker
  output_fields = [ ]
  __fields__ = ['maxRecords', 'code', 'descr',
                'serial', 'bytesPerRecord', 'retries', 'params']
  def __init__(self, **kwds):
    for k in self.__fields__:
      value = kwds.get(k, getattr(self, k))
      setattr(self, k, value)
    self.allocateRawData( )
    self.data = bytearray( )
    self.name = self.log_name( )
    self.checker = self.Validator(self, required=self.output_fields)

  def log_name(self, prefix=''):
    return prefix + '{}.data'.format(self.__class__.__name__)

  def save(self, prefix=''):
    name = '{}'.format(self.log_name(prefix))
    handle = open(name, 'wb')
    handle.write(self.data)
    handle.close( )

  def __str__(self):
    if self.responded:
      return '{}:size[{}]:data:{}'.format(self.__class__.__name__,
                                          self.size, repr(self.getData( )))
    return '{}:data:unknown'.format(self.__class__.__name__)

  def __repr__(self):
    return '<{0}>'.format( self)

  def validate (self, data):
    return True
  def check_output (self, data):
    return self.checker(data)

  def getData(self):
    return self.data

  def allocateRawData(self):
    self.size = self.bytesPerRecord * self.maxRecords

  def format(self):
    params = self.params
    code   = self.code
    maxRetries = self.retries
    serial = list(bytearray(self.serial.decode('hex')))
    paramsCount = len(params)
    head   = [ 1, 0, 167, 1 ]
    # serial
    packet = head + serial
    # paramCount 2 bytes
    packet.extend( [ (0x80 | lib.HighByte(paramsCount)),
                             lib.LowByte(paramsCount) ] )
    # not sure what this byte means
    button = 0
    # special case command 93
    if code == 93:
      button = 85
    packet.append(button)
    packet.append(maxRetries)
    # how many packets/frames/pages/flows will this take?
    responseSize = self.calcRecordsRequired()
    # really only 1 or 2?
    pages = responseSize
    if responseSize > 1:
      pages = 2
    packet.append(pages)
    packet.append(0)
    # command code goes here
    packet.append(code)
    packet.append(CRC8(packet))
    packet.extend(params)
    packet.append(CRC8(params))
    log.debug(packet)
    return bytearray(packet)

  def calcRecordsRequired(self):
    length = self.bytesPerRecord * self.maxRecords
    i = length / 64
    j = length % 64
    if j > 0:
      return i + 1
    return i

class ManualCommand(PumpCommand):
  def __init__(self, **kwds):
    self.name = kwds.get('name', self.__class__.__name__)
    super(type(self), self).__init__(**kwds)
    self.kwds = kwds
    self.name = kwds.get('name', self.__class__.__name__)
  def __str__(self):
    if self.responded:
      return '{}:{}:size[{}]:'.format(self.name, self.kwds,
                                      self.size)
    return '{}:{}:data:unknown'.format(self.name, self.kwds)

  def log_name(self, prefix=''):
    return prefix + '{}.data'.format(self.name)
  def __repr__(self):
    return '<{0}>'.format(self)

  def getData(self):
    return self.hexdump( )

class PowerControl(PumpCommand):
  """
    >>> PowerControl(serial='665455').format() == PowerControl._test_ok
    True
  """
  _test_ok = bytearray( [ 0x01, 0x00, 0xA7, 0x01, 0x66, 0x54, 0x55, 0x80,
                          0x02, 0x55, 0x00, 0x00, 0x00, 0x5D, 0xE6, 0x01,
                          0x0A, 0xA2 ] )
  code = 93
  descr = "RF Power On"
  params = [ 0x01, 0x0A ]
  retries = 0
  maxRecords = 0
  #timeout = 1
  # effectTime = 7
  effectTime = 12
  def __init__(self, minutes=None, **kwds):
    if minutes is not None:
      self.minutes = int(minutes)
      kwds['params'] = [ 0x01, self.minutes ]
    super(PowerControl, self).__init__(**kwds)

class PowerControlOff(PowerControl):
  """
  Here's an example where arguments clearly modify behavior.
  """
  params = [ 0x00, 0x00 ]

# MMPump???/	CMD_????????	69	0x45	('E')	??
class PumpExperiment_OP69 (PumpCommand):
  code = 69

# MMPump???/	CMD_????????	70	0x46	('F')	??
class PumpExperiment_OP70 (PumpCommand):
  code = 70

# MMPump???/	CMD_????????	71	0x47	('G')	??
class PumpExperiment_OP71 (PumpCommand):
  code = 71

# MMPump???/	CMD_????????	72	0x48	('H')	??
class PumpExperiment_OP72 (PumpCommand):
  code = 72

# MMPump???/	CMD_????????	73	0x49	('I')	??
class PumpExperiment_OP73 (PumpCommand):
  code = 73

# MMPump???/	SelectBasalProfile	74	0x4a	('J')	OK
class SelectBasalProfile (PumpCommand):
  code = 74

class SelectBasalProfileSTD (SelectBasalProfile):
  params = [ 0 ]

class SelectBasalProfileA (SelectBasalProfile):
  params = [ 1 ]

class SelectBasalProfileB (SelectBasalProfile):
  params = [ 2 ]

# MMPump???/	CMD_????????	75	0x4b	('K')	??
class PumpExperiment_OP75 (PumpCommand):
  code = 75

class TempBasal(PumpCommand):
  """

  """

  code = 76
  descr = "Set temp basal"
  params = [ 0x00, 0x00, 0x00 ]
  retries = 0
  #maxRecords = 0
  #timeout = 1

  def getData(self):
    status = { 0: 'absolute' }
    received = True if (len(self.data) > 0 and self.data[0] is 0) else False
    return dict(recieved=received, temp=status.get(self.params[0], 'percent'))
  @classmethod
  def Program (klass, rate=None, duration=None, temp=None, **kwds):
    assert duration % 30 is 0, "duration {0} is not a whole multiple of 30".format(duration)
    assert temp in [ 'percent', 'absolute' ], "temp field <{0}> should be one of {1}".format(temp, ['percent', 'absolute' ])
    if temp in [ 'percent' ]:
      return TempBasalPercent(params=klass.format_percent_params(rate, duration), **kwds)

    return klass(params=klass.format_params(rate, duration), **kwds)
  @classmethod
  def format_percent_params (klass, rate, duration):
    duration = int(duration / 30)
    rate = int(rate)
    params = [rate, duration]
    return params

  @classmethod
  def format_params (klass, rate, duration):
    duration = duration / 30
    rate = int(rate / 0.025)
    params = [0x00, rate, duration]
    return params



class SetSuspend(PumpCommand):
  code = 77
  descr = "Set Pump Suspend/Resume status"
  params = [ ]
  retries = 2
  maxRecords = 1
  def getData(self):
    status = { 0: 'resumed', 1: 'suspended' }
    received = True if self.data[0] is 0 else False
    return dict(recieved=received, status=status.get(self.params[0]))

class PumpSuspend(SetSuspend):
  descr = "Suspend pump"
  params = [ 1 ]

class PumpResume(SetSuspend):
  descr = "Resume pump (cancel suspend)"
  params = [ 0 ]

class SetAutoOff (PumpCommand):
  code = 78
  maxRecords = 0

class SetEnabledEasyBolus (PumpCommand):
  code = 79
  maxRecords = 0

class SetBasalType (PumpCommand):
  code = 104

class TempBasalPercent (TempBasal):
  """

  """

  code = 105
  descr = "Set temp basal by percent"
  params = [ 0x00, 0x00 ]
  retries = 0
  #maxRecords = 0
  #timeout = 1

class KeypadPush(PumpCommand):
  code = 91
  descr = "Press buttons on the keypad"
  params = [ ]
  retries = 1
  maxRecords = 0

  @classmethod
  def ACT(klass, **kwds):
    return klass(params=[0x02], **kwds)

  @classmethod
  def ESC(klass, **kwds):
    return klass(params=[0x01], **kwds)

  @classmethod
  def DOWN(klass, **kwds):
    return klass(params=[0x04], **kwds)

  @classmethod
  def UP(klass, **kwds):
    return klass(params=[0x03], **kwds)

  @classmethod
  def EASY(klass, **kwds):
    return klass(params=[0x00], **kwds)

def PushACT (**kwds):
  return KeypadPush.ACT(**kwds)

def PushESC (**kwds):
  return KeypadPush.ESC(**kwds)

def PushDOWN (**kwds):
  return KeypadPush.DOWN(**kwds)

def PushUP (**kwds):
  return KeypadPush.UP(**kwds)

def PushEASY (**kwds):
  return KeypadPush.EASY(**kwds)


class ReadErrorStatus508 (PumpCommand):
  """

  """
  code = 38
  descr = "error status"
  params = [ ]

class ReadBolusHistory (PumpCommand):
  """

  """
  code = 39
  descr = "bolus history"
  params = [ ]

class ReadDailyTotals (PumpCommand):
  """

  """
  code = 40
  descr = "..."
  params = [ ]

class ReadPrimeBoluses (PumpCommand):
  """

  """
  code = 41
  descr = "..."
  params = [ ]

class ReadAlarms (PumpCommand):
  """

  """
  code = 42
  descr = "..."
  params = [ ]

class ReadProfileSets (PumpCommand):
  """

  """
  code = 43
  descr = "..."
  params = [ ]

class ReadUserEvents (PumpCommand):
  """

  """
  code = 44
  descr = "..."
  params = [ ]

class ReadRemoteControlID (PumpCommand):
  """

  """
  code = 46
  descr = "..."
  params = [ ]

class Read128KMem (PumpCommand):
  """

  """
  code = 55
  descr = "..."
  params = [ ]

class Read256KMem (PumpCommand):
  """

  """
  code = 56
  descr = "..."
  params = [ ]

class Bolus (PumpCommand):
  """
  Bolus some insulin.

  XXX: Be careful please.
  Best trying this not connected to the pump until you trust it.
  """
  code = 66
  descr = "Bolus"
  params = [ ]
  def getData(self):
    received = True if self.data[0] is 0x0c else False
    return dict(recieved=received, _type='BolusRequest')


class ReadErrorStatus(PumpCommand):
  """
    >>> ReadErrorStatus(serial='665455').format() == ReadErrorStatus._test_ok
    True
  """
  _test_ok = bytearray([ 0x01, 0x00, 0xA7, 0x01, 0x66, 0x54, 0x55, 0x80,
                         0x00, 0x00, 0x02, 0x01, 0x00, 0x75, 0xD7, 0x00 ])
  code = 117
  descr = "Read Error Status any current alarms set?"
  params = [ ]
  retries = 2
  maxRecords = 1

class ReadHistoryData(PumpCommand):
  """
    >>> ReadHistoryData(serial='208850', params=[ 0x03 ]).format() == ReadHistoryData._test_ok
    True
    >>> ReadHistoryData(params=[ 0x01 ]).params
    [1]
    >>> ReadHistoryData(params=[ 0x02 ]).params
    [2]
    >>> ReadHistoryData(params=[ 0x03 ]).params
    [3]
    >>> ReadHistoryData(page=0x01).params
    [1]
    >>> ReadHistoryData(page=0x02).params
    [2]
    >>> ReadHistoryData(page=0x03).params
    [3]
  """
  __fields__ = PumpCommand.__fields__ + ['page']
  _test_ok = bytearray([ 0x01, 0x00, 0xA7, 0x01, 0x20, 0x88, 0x50, 0x80, 0x01, 0x00, 0x02, 0x02, 0x00, 0x80, 0x9B, 0x03, 0x36, ])

  page = None
  def __init__(self, page=None, **kwds):
    if page is None and kwds.get('params', [ ]):
      page = kwds.pop('params')[0] or 0

    if page is not None:
      self.page = int(page)
      kwds['params'] = [ self.page ]
    super(ReadHistoryData, self).__init__(**kwds)

  def log_name(self, prefix=''):
    return prefix + '{}-page-{}.data'.format(self.__class__.__name__, self.page)

  def __str__(self):
    base = ''.join([ self.__class__.__name__,
                    ':size[%s]:' % self.size,
                    '[page][%s]' % self.page ])
    return '{}:data[{}]:'.format(base, len(self.data))

  def done(self):
    eod = False
    found = len(self.data or [ ])
    expect = int(self.maxRecords * self.bytesPerRecord)
    expect_crc = CRC8(self.data[:-1])
    expect_size = "size check found[{}] expected[{}]".format(found, expect)
    found_crc = 0
    if self.responded and len(self.data) > 5:
      found_crc = self.data[-1]
      self.eod  = eod = (self.data[5] & 0x80) > 0
    explain_crc = "CRC ACK check found[{}] expected[{}]".format(found_crc, expect_crc)
    is_eod = 'and has eod set? %s' % (eod)
    log.info("%s:download:done %s:%s:%s" % (self, expect_size, explain_crc, is_eod))
    return found >= expect

  def respond(self, raw):
    log.info('{} extending original {} with found {}'.format(str(self), len(self.data), len(raw)))
    if len(raw) == self.size:
      log.info('{} download respond replace original {} with found {}'.format(str(self), len(self.data), len(raw)))
      self.data = raw
    elif len(self.data) == self.size:
      log.info('{} download respond original {}, XXX IGNORE found {}'.format(str(self), len(self.data), len(raw)))
      pass
    else:
      log.info('{} download respond extend original {} with found {}'.format(str(self), len(self.data), len(raw)))
      self.data.extend(raw)
    self.responded = True

  code = 128
  descr = "Read History Data"
  params = [ ]
  retries = 2
  maxRecords = 16
  effectTime = .100
  data = bytearray( )

  def getData(self):
    data = self.data
    # log.info("XXX: READ HISTORY DATA!!:\n%s" % lib.hexdump(data))
    return self.hexdump( )

class ReadCurPageNumber(PumpCommand):
  """
  """

  code = 157
  descr = "Read Cur Page Number"
  params = [ ]
  retries = 2
  maxRecords = 1
  pages = 'unknown'

  def __str__(self):
    return ':pages:'.join([self.__class__.__name__, str(self.pages) ])

  def respond(self, data):
    self.data = data
    self.pages = self.getData( )
    self.responded = True
  def getData(self):
    data = self.data
    log.info("XXX: READ cur page number:\n%s" % lib.hexdump(data))
    # MM12 does not support this command, but has 31 pages
    # Thanks to @amazaheri
    page = 32
    if len(data) == 1:
      return int(data[0])
    if len(data) > 3:
      page = lib.BangLong(data[0:4])
    # https://bitbucket.org/bewest/carelink/src/419fbf23495a/ddmsDTWApplet.src/minimed/ddms/deviceportreader/MMX15.java#cl-157
    if page <= 0 or page > 36:
      page = 36
    return page


# MMX22/	CMD_READ_CURRENT_GLUCOSE_HISTORY_PAGE_NUMBER	205	0xcd	('\xcd')	OK
class ReadCurGlucosePageNumber(PumpCommand):
  """
  """

  code = 205
  descr = "Read Cur Glucose Page Number"
  params = [ ]
  retries = 2
  maxRecords = 1

  def getData(self):
    data = self.data
    log.info("XXX: READ cur page number:\n%s" % lib.hexdump(data))
    if len(data) == 1:
      return int(data[0])
    return dict(page= lib.BangLong(data[0:4]), glucose=data[5], isig=data[7])


class ReadRTC(PumpCommand):
  """
  """

  code = 112
  descr = "Read RTC"
  params = [ ]
  retries = 2
  maxRecords = 1


  def getData(self):
    data = self.data
    d = {
      'hour'  : int(data[0]),
      'minute': int(data[1]),
      'second': int(data[2]),
      # XXX
      'year'  : lib.BangInt([data[3], data[4]]),
      'month' : int(data[5]),
      'day'   : int(data[6]),
    }
    return "{year:#04}-{month:#02}-{day:#02}T{hour:#02}:{minute:#02}:{second:#02}".format(**d)

class SetRTC (PumpCommand):
  """
  Set clock
  """
  code = 64
  descr = "Set RTC"
  retries = 2
  maxRecords = 0

  __fields__ = PumpCommand.__fields__ + ['clock']
  def __init__(self, clock=None, **kwds):
    params = kwds.get('params', [ ])
    self.clock = kwds.get('clock', None)
    if len(params) == 0:
      params.extend(SetRTC.fmt_datetime(clock))

    kwds['params'] = params
    super(SetRTC, self).__init__(**kwds)
  @classmethod
  def fmt_datetime (klass, dt):
    return [dt.hour, dt.minute, dt.second, lib.HighByte(dt.year), lib.LowByte(dt.year), dt.month, dt.day]

class ReadPumpID(PumpCommand):
  """
  """

  code = 113
  descr = "Read Pump ID"
  params = [ ]
  retries = 2
  maxRecords = 1

  def getData(self):
    data = self.data
    return str(data[0:6])

class ReadBatteryStatus(PumpCommand):
  """
  """

  code = 114
  descr = "Read Battery Status"
  params = [ ]
  retries = 2
  maxRecords = 1

  def getData(self):
    data = self.data
    bd = bytearray(data)
    volt = lib.BangInt((bd[1], bd[2]))
    indicator = bd[0]
    battery = {'status': {0: 'normal', 1: 'low'}[indicator], 'voltage': volt/100.0 }
    return battery


class ReadFirmwareVersion(PumpCommand):
  """
  """

  code = 116
  descr = "Read Firmware Version"
  params = [ ]
  retries = 2
  maxRecords = 1

  def getData(self):
    data = self.data
    log.debug("READ FIRMWARE HEX:\n%s" % lib.hexdump(data))
    return str(data.split( chr(0x0b) )[0]).strip( )

class ReadRemainingInsulin(PumpCommand):
  """
  """

  code = 115
  descr = "Read Remaining Insulin"
  params = [ ]
  retries = 2
  maxRecords = 1
  basalStrokes = 10.0
  startByte = 0
  endByte = 2

  def getData(self):
    data = self.data
    log.info("READ remaining insulin:\n%s" % lib.hexdump(data))
    return lib.BangInt(data[self.startByte:self.endByte])/self.basalStrokes


class ReadRemainingInsulin523(ReadRemainingInsulin):
  """
  """

  basalStrokes = 40.0
  startByte = 2
  endByte = 4


class ReadBasalTemp508 (PumpCommand):
  """
  """

  code = 64
  descr = "Read Temp Basal 508 (old)"
  params = [ ]
  retries = 2
  maxRecords = 1

  def getData(self):
    data = self.data
    rate = lib.BangInt(data[2:4])/40.0
    duration = lib.BangInt(data[4:6])
    log.info("READ temporary basal:\n%s" % lib.hexdump(data))
    return { 'rate': rate, 'duration': duration }


class ReadTodayTotals508 (PumpCommand):
  """
  """

  code = 65
  descr = "Read Totals Today"
  params = [ ]
  retries = 2
  maxRecords = 1

  def getData(self):
    data = self.data
    log.info("READ totals today:\n%s" % lib.hexdump(data))
    totals = {
      'today': lib.BangInt(data[0:2]) / 10.0,
      'yesterday': lib.BangInt(data[2:4]) / 10.0
    }
    return totals

# MMPump511/	ReadTotalsToday	121	0x79	('y')	OK
class ReadTotalsToday(PumpCommand):
  """
  """

  code = 121
  descr = "Read Totals Today"
  params = [ ]
  retries = 2
  maxRecords = 1

  def getData(self):
    data = self.data
    log.info("READ totals today:\n%s" % lib.hexdump(data))
    totals = {
      'today': lib.BangInt(data[0:2]) / 10.0,
      'yesterday': lib.BangInt(data[2:4]) / 10.0
    }
    return totals

# MMPump511/	ReadProfiles_STD	122	0x7a	('z')	OK
class ReadProfiles511_STD (PumpCommand):
  code = 122
# MMPump511/	ReadProfiles_A	123	0x7b	('{')	??
class ReadProfiles511_A (PumpCommand):
  code = 123
# MMPump511/	ReadProfiles_B	124	0x7c	('|')	??
class ReadProfiles511_B (PumpCommand):
  code = 124
# MMPump???/	CMD_?????	125	0x7d	('}')	??
class Model511_ExperimentOP125 (PumpCommand):
  code = 125
# MMPump???/	CMD_?????	126	0x7e	('~')	??
class Model511_ExperimentOP126 (PumpCommand):
  code = 126
# MMPump511/	ReadSettings	127	0x7f	DEL
class ReadSettings511 (PumpCommand):
  code = 127

# MMX11/	CMD_ENABLE_DISABLE_DETAIL_TRACE	160	0x9f	('\x9f')	??
class PumpTraceSelect (PumpCommand):
  code = 160

class PumpEnableDetailTrace (PumpTraceSelect):
  params = [ 1 ]

class PumpDisableDetailTrace (PumpTraceSelect):
  params = [ 0 ]

class Experiment_OP161 (PumpCommand):
  code = 161

class Experiment_OP162 (PumpCommand):
  code = 162

# MMPump511/	ReadPumpTrace	163	0xa3	('\xa3')	??
class ReadPumpTrace (PumpCommand):
  code = 163
  maxRecords = 16
# MMPump511/	ReadDetailTrace	164	0xa4	('\xa4')	??
class ReadDetailTrace (PumpCommand):
  code = 164
  maxRecords = 16

# MMPump11??/	CMD_????????????	165	0xa5	0xa5	??
class Model511_Experiment_OP165 (PumpCommand):
  code = 165

# MMPump511/	ReadNewTraceAlarm	166	0xa6	('\xa6')	??
class ReadNewTraceAlarm (PumpCommand):
  code = 166
  maxRecords = 16

# MMPump511/	ReadOldTraceAlarm	167	0xa7	('\xa7')	??
class ReadOldTraceAlarm (PumpCommand):
  maxRecords = 16
  code = 167

# MMPump???/	CMD_?????	36	0x24	('$')	??
class PumpExperimentSelfCheck_OP36 (PumpCommand):
  code = 36

# MMX22/	CMD_WRITE_GLUCOSE_HISTORY_TIMESTAMP	40	0x28	('(')	??
class WriteGlucoseHistoryTimestamp (PumpCommand):
  code = 40

class ReadRadioCtrlACL(PumpCommand):
  """
  """

  code = 118
  descr = "Read Radio ACL"
  params = [ ]
  retries = 2
  maxRecords = 1

  def getData(self):
    data = self.data
    ids = [ ]
    ids.append( str(data[0:6]) )
    ids.append( str(data[6:12]) )
    ids.append( str(data[12:18]) )
    log.info("READ radio ACL:\n%s" % lib.hexdump(data))
    return ids

class Model511_Experiment_OP119 (PumpCommand):
  code = 119

class Model511_Experiment_OP120 (PumpCommand):
  code = 120

class Model511_Experiment_OP121 (PumpCommand):
  code = 121

class Model511_Experiment_OP122 (PumpCommand):
  code = 122

class Model511_Experiment_OP123 (PumpCommand):
  code = 123

class Model511_Experiment_OP124 (PumpCommand):
  code = 124

class Model511_Experiment_OP125 (PumpCommand):
  code = 125

class Model511_Experiment_OP126 (PumpCommand):
  code = 126

class Model511_Experiment_OP127 (PumpCommand):
  code = 127

class Model511_Experiment_OP128 (PumpCommand):
  code = 128

class Model511_Experiment_OP129 (PumpCommand):
  code = 129

class Model511_Experiment_OP130 (PumpCommand):
  code = 130

# MMPump512/	CMD_READ_LANGUAGE	134	0x86	('\x86')	??
class ReadLanguage (PumpCommand):
  code = 134
# MMPump512/	CMD_READ_BOLUS_WIZARD_SETUP_STATUS	135	0x87	('\x87')	??
class ReadBolusWizardSetupStatus (PumpCommand):
  code = 135

# MMPump512/	CMD_READ_CARB_UNITS	136	0x88	('\x88')	OK
class ReadCarbUnits (PumpCommand):
  code = 136
  def getData (self):
    labels = { 1 : 'grams', 2: 'exchanges' }
    return dict(carb_units=labels.get(self.data[0], self.data[0]))

# MMPump512/	CMD_READ_BG_UNITS	137	0x89	('\x89')	??
class ReadBGUnits (PumpCommand):
  code = 137
  def getData (self):
    labels = { 1 : 'mg/dL', 2: 'mmol/L' }
    return dict(bg_units=labels.get(self.data[0], self.data[0]))

# MMPump512/	CMD_READ_CARB_RATIOS	138	0x8a	('\x8a')	OK
class ReadCarbRatios512 (PumpCommand):
  code = 138
  output_fields = ['units', 'schedule' ]
  def getData (self):
    # return self.model.decode_carb_ratios(self.data[:])
    units = self.data[0]
    labels = { 1 : 'grams', 2: 'exchanges' }
    fixed = self.data[1]
    data = self.data[1:1+(8 *2)]
    return dict(schedule=self.decode_ratios(data[0:], units=units), units=labels.get(units), first=self.data[0], raw=' '.join('0x{:02x}'.format(x) for x in self.data))

  item_size = 2
  num_items = 8
  @classmethod
  def decode_ratios (klass, data, units=0):
    data = data[0:(8 *2)]
    schedule = [ ]
    for x in range(len(data)/ 2):
      start = x * 2
      end = start + 2
      (i, r) = data[start:end]
      if x > 0 and i == 0:
        break
      ratio = int(r)
      if units == 2:
        ratio = r / 10.0
      schedule.append(dict(x=x, i=i, start=lib.basal_time(i), offset=i*30, ratio=ratio, r=r))
    return schedule

class ReadCarbRatios (PumpCommand):
  code = 138
  item_size = 3
  num_items = 8
  output_fields = ['units', 'schedule' ]
  def getData (self):
    units = self.data[0]
    labels = { 1 : 'grams', 2: 'exchanges' }
    fixed = self.data[1]
    data = self.data[2:2+(fixed *3)]
    return dict(schedule=self.decode_ratios(data, units=units), units=labels.get(units), first=self.data[0])

  @classmethod
  def decode_ratios (klass, data, units=0):
    schedule = [ ]
    for x in range(len(data)/ 3):
      start = x * 3
      end = start + 3
      (i, q, r) = data[start:end]
      if x > 0 and i == 0:
        break
      ratio = r/10.0
      if q:
        ratio = lib.BangInt([q, r]) / 1000.0
      schedule.append(dict(x=x, i=i, start=lib.basal_time(i), offset=i*30, q=q, ratio=ratio, r=r))
    return schedule

# MMPump512/	CMD_READ_INSULIN_SENSITIVITIES	139	0x8b	('\x8b')	OK
class ReadInsulinSensitivities (PumpCommand):
  code = 139
  resp_1 = bytearray(b'\x01\x00-\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00')
  resp_uk_1 = bytearray(str("""
02001600000000
00000000000000
00000000000000
00000000000000
00000000000000
00000000000000
00000000000000
00000000000000
""".strip( ).replace("\n", "").decode('hex')))

  output_fields = ['units', 'sensitivities' ]
  UNITS = {
    1: 'mg/dL',
    2: 'mmol/L'
  }
  def getData (self):
    # isFast = data[17] == 0
    isFast = self.data[0] is 1
    units = self.data[0]
    data = self.data[1:1+16]
    schedule = [ ]
    for x in range(8):
      start = x * 2
      end = start + 2
      (i, sensitivity) = data[start:end]
      if x > 0 and i == 0:
        break
      if units == 2:
        sensitivity = sensitivity / 10.0
      schedule.append(dict(x=x, i=i, start=lib.basal_time(i), offset=i*30, sensitivity=sensitivity))
    labels = { True: 'Fast', False: 'Regular' }
    return dict(sensitivities=schedule, first=self.data[0], units=self.UNITS.get(units))

# MMPump512/	CMD_READ_BG_TARGETS	140	0x8c	('\x8c')	??
class ReadBGTargets (PumpCommand):
  code = 140
class ReadBGTargets515 (PumpCommand):
  code = 159
  output_fields = ['units', 'targets' ]
  def getData (self):
    units = self.data[0]
    labels = { 1 : 'mg/dL', 2: 'mmol/L' }
    data = self.data[1:1+24]
    schedule = [ ]
    for x in range(8):
      start = x * 3
      end = start + 3
      (i, low, high) = data[start:end]
      if x > 0 and i == 0:
        break
      if units is 2:
        low = low / 10.0
        high = high / 10.0
      schedule.append(dict(x=x, i=i, start=lib.basal_time(i), offset=i*30, low=low, high=high))
    return dict(targets=schedule, units=labels.get(units), first=self.data[0], raw=' '.join('0x{:02x}'.format(x) for x in self.data))

# MMPump512/	CMD_READ_BG_ALARM_CLOCKS	142	0x8e	('\x8e')	??
class ReadBGAlarmCLocks (PumpCommand):
  code = 142
# MMPump512/	CMD_READ_RESERVOIR_WARNING	143	0x8f	('\x8f')	??
class ReadReservoirWarning (PumpCommand):
  code = 143
# MMPump512/	CMD_READ_BG_REMINDER_ENABLE	144	0x90	('\x90')	??
class ReadBGReminderEnable (PumpCommand):
  code = 144
# MMPump512/	CMD_READ_SETTINGS	145	0x91	('\x91')	??
class ReadSettings512 (PumpCommand):
  code = 145
# MMPump512/	CMD_READ_STD_PROFILES	146	0x92	('\x92')	??
class ReadProfile_STD512 (PumpCommand):
  """
    >>> import json
    >>> schedule = ReadProfile_STD512.decode(ReadProfile_STD512._test_result_1)
    >>> len(schedule)
    4
    >>> print json.dumps(schedule[0])
    {"i": 0, "start": "00:00:00", "rate": 0.8, "minutes": 0}
    >>> print json.dumps(schedule[1])
    {"i": 1, "start": "06:30:00", "rate": 0.9500000000000001, "minutes": 390}
    >>> print json.dumps(schedule[2])
    {"i": 2, "start": "09:30:00", "rate": 1.1, "minutes": 570}
    >>> print json.dumps(schedule[3])
    {"i": 3, "start": "14:00:00", "rate": 0.9500000000000001, "minutes": 840}

  """
  _test_result_1 = bytearray([
    32, 0, 0,
    38, 0, 13,
    44, 0, 19,
    38, 0, 28,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0
  ])
  _test_schedule = {'total': 22.50, 'schedule': [
    { 'start': '12:00A', 'rate': 0.80 },
    { 'start': '6:30A', 'rate': 0.95 },
    { 'start': '9:30A', 'rate': 1.10 },
    { 'start': '2:00P', 'rate': 0.95 },
  ]}
  code = 146
  maxRecords = 2
  output_fields = [ ]
  def validate (self, data):
    i = 0
    valid = True
    last = None
    for profile in data:
      start = str(lib.basal_time(profile['minutes']/30))
      if 'rate' in profile and profile['i'] == i and start == profile['start']:
        if i == 0:
          if profile['minutes'] != 0:
            template = "{name} first scheduled item should be 00:00:00."
            msg = template.format(name=self.__class__.__name__)
            msg = "%s\n%s" % (msg, profile)
            raise BadResponse(msg)
        if last and profile['minutes'] <= last['minutes']:
          template = "{name} next scheduled item occurs before previous"
          msg = template.format(name=self.__class__.__name__)
          msg = "%s\n%s" % (msg, profile)
          raise BadResponse(msg)
      else:
        bad_profile = """Current profile: %s""" % (profile)
        template = """{bad_profile} Found in response to {name}
          i: {i} matches? {matches_i}
          our calcstart: {start}
          profile start: {profile_start}
          has a rate: {has_rate}
          start matches: {matches_start}
        """
        raise BadResponse(template.format(bad_profile=bad_profile
                  , name=self.__class__.__name__
                  , start=start
                  , profile_start=profile['start']
                  , has_rate= 'rate' in profile
                  , matches_i= profile['i'] == i
                  , matches_start=  start == profile['start']
                  , i=i))
        valid = False
      last = profile
      i = i + 1
    return True
  @staticmethod
  def decode (data):
    i = 0
    schedule = [ ]
    end = [ 0, 0, 0 ]
    none = [ 0, 0, 0x3F ]
    for i in xrange(len(data)/3):
      off = i*3
      r, z, m = data[off : off + 3]
      if [r,z,m] in [end, none]:
        break
      schedule.append(dict(i=i, minutes=m*30, start=str(lib.basal_time(m)), rate=r*.025))
    return schedule
  def getData (self):
    return self.decode(self.data)
# MMPump512/	CMD_READ_A_PROFILES	147	0x93	('\x93')	OK
class ReadProfile_A512 (ReadProfile_STD512):
  code = 147
# MMPump512/	CMD_READ_B_PROFILES	148	0x94	('\x94')	OK
class ReadProfile_B512 (ReadProfile_STD512):
  code = 148
# MMPump512/	CMD_READ_LOGIC_LINK_IDS	149	0x95	('\x95')	OK
class ReadLogicLinkIDS (PumpCommand):
  code = 149

# MMPump512??/	CMD_????????????????	150	0x96	('\x96')	??
class Model512Experiment_OP150 (PumpCommand):
  code = 150

# MMPump512/	CMD_READ_BG_ALARM_ENABLE	151	0x97	('\x97')	??
class ReadBGAlarmEnable (PumpCommand):
  code = 151

# MMPump512/	CMD_READ_TEMP_BASAL	152	0x98	('\x98')	OK
class ReadBasalTemp(PumpCommand):
  """
  MM511 - 120
  MM512 and up - opcode 152
  # strokes per basalunit = 40 - mm12, 10 in mm11
  """

  code = 152
  descr = "Read Temp Basal"
  params = [ ]
  retries = 2
  maxRecords = 1

  def getData(self):
    data = self.data
    temp = { 0: 'absolute', 1: 'percent' }[self.data[0]]
    status = dict(temp=temp)
    if temp is 'absolute':
      rate = lib.BangInt(data[2:4])/40.0
      duration = lib.BangInt(data[4:6])
      status.update(rate=rate, duration=duration)
    if temp is 'percent':
      rate = int(data[1])
      duration = lib.BangInt(data[4:6])
      status.update(rate=rate, duration=duration)
    log.info("READ temporary basal:\n%s" % lib.hexdump(data))
    return status

# MMGuardian3/	CMD_READ_SENSOR_SETTINGS	207	0xcf	('\xcf')	??
class GuardianSensorSettings (PumpCommand):
  code = 207
# MMGuardian3/	CMD_READ_SENSOR_PREDICTIVE_ALERTS	209	0xd1	('\xd1')	??
class GuardianSensorSettings (PumpCommand):
  code = 209
# MMGuardian3/	CMD_READ_SENSOR_DEMO_AND_GRAPH_TIMEOUT	210	0xd2	('\xd2')	??
class GuardianSensorDemoGraphTimeout (PumpCommand):
  code = 210
# MMGuardian3/	CMD_READ_SENSOR_ALARM_SILENCE	211	0xd3	('\xd3')	??
class GuardianSensorAlarmSilence (PumpCommand):
  code = 211
# MMGuardian3/	CMD_READ_SENSOR_RATE_OF_CHANGE_ALERTS	212	0xd4	('\xd4')	??
class GuardianSensorRateChangeAlerts (PumpCommand):
  code = 212

class ReadSettings(PumpCommand):
  """
  XXX: changed in MM512 to 192

  """

  code = 192
  descr = "Read Settings"
  params = [ ]
  retries = 2
  maxRecords = 1

  output_fields = ['maxBolus', 'maxBasal', 'insulin_action_curve' ]
  byte_map = {

  }

  def alarm(self, alarm):
    d = { 'volume': alarm, 'mode': 2 }
    if alarm == 4:
      d = { 'volume': -1, 'mode': 1 }
    return d
  def temp_basal_type(self, data):
    temp = { 'type': data[0] == 1 and "Percent" or "Units/hour",
             'percent': data[1]
           }
    return temp
  def getData(self):
    data = self.data
    log.info("READ pump settings:\n%s" % lib.hexdump(data))

    if len(data) < 2:
      log.info("pump settings: unsupported version, sorry")
      return data

    auto_off_duration_hrs = data[0]
    alarm = self.alarm(data[1])
    audio_bolus_enable = data[2] == 1
    audio_bolus_size = 0
    if audio_bolus_enable:
      audio_bolus_size = data[3] / 10.0
    variable_bolus_enable = data[4] == 1
    #MM23 is different
    maxBolus = data[5]/ 10.0
    # MM512 and up
    maxBasal = lib.BangInt(data[6:8]) / 40.0
    timeformat = data[8]
    insulinConcentration = {0: 100, 1: 50}[data[9]]
    patterns_enabled = data[10] == 1
    selected_pattern = data[11]
    rf_enable = data[12] == 1
    block_enable = data[13] == 1
    temp_basal = self.temp_basal_type(data[14:16])
    paradigm_enabled = data[16]
    """
    # MM12
    insulin_action_type = data[17] == 0 and 'Fast' or 'Regular'
    """
    #MM15
    # insulin_action_type = data[17]
    insulin_action_curve = data[17]
    low_reservoir_warn_type = data[18]
    low_reservoir_warn_point = data[19]
    keypad_lock_status = data[20]

    values = locals( )
    # safety
    values.pop('self')
    values.pop('data')

    return values

class ReadSettings523(ReadSettings):

  def getData(self):
    values = super(ReadSettings523, self).getData()
    data = self.data

    values['maxBasal'] = lib.BangInt(data[7:9]) / 40.0
    values['maxBolus'] = data[6]/ 10.0

    return values

# MMX15/	CMD_READ_SAVED_SETTINGS_DATE	193	0xc1	('\xc1')	??
class ReadSavedSettingsDate (PumpCommand):
  code = 193

class ReadContrast(PumpCommand):
  """
  """

  code = 195
  descr = "Read Contrast"
  params = [ ]
  retries = 2
  maxRecords = 1

  def getData(self):
    data = self.data
    log.info("READ contrast:\n%s" % lib.hexdump(data))
    return data



# MMX15/	CMD_READ_BOLUS_REMINDER_ENABLE	197	0xc5	('\xc5')	??
class ReadBolusReminderEnable (PumpCommand):
  code = 197

# MMX15/	CMD_READ_BOLUS_REMINDERS	198	0xc6	('\xc6')	??
class ReadBolusReminders (PumpCommand):
  code = 198

# MMX15/	CMD_READ_FACTORY_PARAMETERS	199	0xc7	('\xc7')	??
class ReadFactoryParameters (PumpCommand):
  code = 199

class ReadPumpStatus(PumpCommand):
  """
  """


  code = 206
  descr = "Read Pump Status"
  params = [ ]
  retries = 2
  maxRecords = 1
  def getData(self):
    data = self.data
    normal = { 03: 'normal' }
    status = { 'status': normal.get(data[0], 'error'),
               'bolusing': data[1] == 1,
               'suspended': data[2] == 1
             }
    return status

class ReadPumpState(PumpCommand):
  """
    >>> ReadPumpState(serial='665455').format() == ReadPumpState._test_ok
    True
  """
  _test_ok = bytearray([ 0x01, 0x00, 0xA7, 0x01, 0x66, 0x54, 0x55, 0x80,
                         0x00, 0x00, 0x02, 0x01, 0x00, 0x83, 0x2E, 0x00 ])

  code = 131
  descr = "Read Pump State"
  params = [ ]
  retries = 2
  maxRecords = 1



# MMX22/	CMD_READ_SENSOR_SETTINGS	153	0x99	('\x99')	??
class ReadSensorSettings (PumpCommand):
  """
  """
  descr = "Read sensor settings"
  code = 153
  params = [ ]
  retries = 2

class ReadSensorHistoryData (ReadHistoryData):
  def __init__(self, page=None, **kwds):
    params = kwds.pop('params', [ ])
    if page is not None:
      params = [ lib.LowByte(page >> 24), lib.LowByte(page >> 16),
                 lib.LowByte(page >>  8), lib.LowByte(page) ]

      self.page = page
    super(ReadSensorHistoryData, self).__init__(params=params, **kwds)
    self.params = params
    self.page = page

# MMX22/	CMD_READ_GLUCOSE_HISTORY	154	0x9a	('\x9a')	??
class ReadGlucoseHistory (ReadSensorHistoryData):
  """
    >>> ReadGlucoseHistory(page=1).params
    [0, 0, 0, 1]
    >>> list(ReadGlucoseHistory(page=1).format( ))
    [1, 0, 167, 1, 32, 136, 80, 128, 4, 0, 2, 2, 0, 154, 34, 0, 0, 0, 1, 155]
    >>> ReadGlucoseHistory(page=2).params
    [0, 0, 0, 2]
    >>> ReadGlucoseHistory(page=3)
    <ReadGlucoseHistory:size[1024]:[page][3]:data[0]:>
    >>> list(ReadGlucoseHistory(page=3).format( ))
    [1, 0, 167, 1, 32, 136, 80, 128, 4, 0, 2, 2, 0, 154, 34, 0, 0, 0, 3, 54]
    >>> ReadGlucoseHistory(page=3).params
    [0, 0, 0, 3]
    >>> ReadGlucoseHistory(params=[1]).params
    [1]
    >>> ReadGlucoseHistory(params=[2]).params
    [2]
    >>> ReadGlucoseHistory(params=[3]).params
    [3]
  """
  descr = "Read glucose history"
  code = 154
  params = [ ]

# MMX22/	CMD_READ_ISIG_HISTORY	155	0x9b	('\x9b')	??
class ReadISIGHistory (ReadSensorHistoryData):
  """
    >>> ReadISIGHistory(page=0).params
    [0, 0, 0, 0]

    >>> ReadISIGHistory(page=1).params
    [0, 0, 0, 1]

    >>> ReadISIGHistory(page=2).params
    [0, 0, 0, 2]

  """
  descr = "Read ISIG history"
  code = 155
  params = [ ]
  maxRecords = 32

# MMX22/	CMD_READ_CALIBRATION_FACTOR	156	0x9c	('\x9c')	??
class ReadCalibrationFactor (PumpCommand):
  """
  """
  code = 156

# MMX23/	CMD_READ_VCNTR_HISTORY	213	0xd5	('\xd5')	??
class ReadVCNTRHistory (ReadSensorHistoryData):
  code = 213

# MMX23/	CMD_READ_OTHER_DEVICES_IDS	240	0xf0	('\xf0')	??
class ReadOtherDevicesIDS (PumpCommand):
  code = 240


class ReadCaptureEventEnabled (PumpCommand):
  code = 241


class ChangeCaptureEventEnable (PumpCommand):
  code = 242
  params = [0]

  def __init__(self, enabled=True, **kwds):
    self.params[0] = int(enabled)
    super(ChangeCaptureEventEnable, self).__init__(**kwds)


class ReadConnectDevicesOtherDevicesStatus (PumpCommand):
  code = 243


class FilterHistory (PumpCommand):
  """

  """
  code = None
  begin = None
  end = None
  __fields__ = PumpCommand.__fields__ + ['begin', 'end']

  def __init__(self, begin=None, end=None, **kwds):
    params = kwds.get('params', [ ])
    if len(params) == 0:
      params.extend(lib.format_filter_date(begin))
      params.extend(lib.format_filter_date(end))

    kwds['params'] = params
    super(FilterHistory, self).__init__(**kwds)

  def getData(self):
    data = self.data
    if len(data) < 4:
      return bytearray(data)
    begin = lib.BangInt(data[0:2])
    end = lib.BangInt(data[2:4])
    return dict(begin=begin, end=end, params=self.params)

  @classmethod
  def ISO (klass, begin=None, end=None, **kwds):
    return klass(begin=lib.parse.date(begin), end=lib.parse.date(end), **kwds)

# MMX22??/	CMD_FILTER_BG	168	0xa8	('\xa8')	??
class FilterGlucoseHistory (FilterHistory):
  """
    >>> FilterGlucoseHistory.ISO(begin='2014-04-13', end='2014-04-14').params
    [7, 222, 4, 13, 7, 222, 4, 14]
  """
  code = 168

# MMX22??/	CMD_FILTER_ISIG	169	0xa9	('\xa9')	??
class FilterISIGHistory (FilterHistory):
  """
    >>> FilterISIGHistory.ISO(begin='2014-04-13', end='2014-04-14').params
    [7, 222, 4, 13, 7, 222, 4, 14]

  """
  code = 169

class TweakAnotherCommand (ManualCommand):
  @classmethod
  def get_kwds (klass, Other, args):
    kwds = { }
    fields = list(set(Other.__fields__) - set(['serial', ]))
    for k in fields:
      value = getattr(args, k, None)
      if value is not None:
        kwds[k] = value
    return kwds

  @classmethod
  def config_argparse (klass, parser):
    parser.add_argument('--params', type=int, action="append",
                        help="parameters to format into sent message"
                       )
    parser.add_argument('--params_hexline', dest='params', type=lib.decode_hexline,
                        help="hex string, parameters to format into sent message"
                        # default=commands.ManualCommand.params
                       )
    parser.add_argument('--descr', type=str,
                        help="Description of command"
                       )
    parser.add_argument('--name', type=str,
                        help="Proposed name of command"
                       )
    parser.add_argument('--save', action="store_true", default=False,
                        help="Save response in a file."
                       )
    parser.add_argument('--effectTime', type=float,
                        help="time to sleep before responding to message, float in seconds"
                       )
    parser.add_argument('--maxRecords', type=int,
                        help="number of frames in a packet composing payload response"
                       )
    parser.add_argument('--bytesPerRecord', type=int,
                        help="bytes per frame"
                       )

    parser.add_argument('--page', type=int,
                        help="Page to fetch (for ReadHistoryData)"
                       )

    parser.add_argument('--begin', type=lib.parse.date,
                        help="begin date for FilterHistory"
                       )
    parser.add_argument('--end', type=lib.parse.date,
                        help="end date for FilterHistory"
                       )


    return parser


class ReadPumpModel(PumpCommand):
  """
    >>> ReadPumpModel(serial='665455').format() == ReadPumpModel._test_ok
    True
  """
  code = 141
  descr = "Read Pump Model Number"
  params = [ ]
  retries = 2
  maxRecords = 1
  _test_ok = bytearray([ 0x01, 0x00, 0xA7, 0x01, 0x66, 0x54, 0x55, 0x80,
                         0x00, 0x00, 0x02, 0x01, 0x00, 0x8D, 0x5B, 0x00 ])

  def getData(self):
    data = self.data
    if len(data) == 0:
      return ''
    length = data[0]
    msg = data[1:1+length]
    self.model = msg
    return str(msg)

def do_commands(device):
  comm = ReadPumpModel( serial=device.serial )
  device.execute(comm)
  log.info('comm:%s:data:%s' % (comm, getattr(comm.getData( ), 'data', None)))
  log.info('REMOTE PUMP MODEL NUMBER: %s' % comm.getData( ))

  log.info("READ RTC")
  comm = ReadRTC( serial=device.serial )
  device.execute(comm)
  log.info('comm:RTC:%s' % (comm.getData( )))

  log.info("READ PUMP ID")
  comm = ReadPumpID( serial=device.serial )
  device.execute(comm)
  log.info('comm:READ PUMP ID: ID: %s' % (comm.getData( )))


  log.info("Battery Status")
  comm = ReadBatteryStatus( serial=device.serial )
  device.execute(comm)
  log.info('comm:READ Battery Status: %r' % (comm.getData( )))

  log.info("Firmware Version")
  comm = ReadFirmwareVersion( serial=device.serial )
  device.execute(comm)
  log.info('comm:READ Firmware Version: %r' % (comm.getData( )))

  log.info("remaining insulin")
  comm = ReadRemainingInsulin( serial=device.serial )
  device.execute(comm)
  log.info('comm:READ Remaining Insulin: %r' % (comm.getData( )))

  log.info("read totals today")
  comm = ReadTotalsToday( serial=device.serial )
  device.execute(comm)
  log.info('comm:READ totals today: %r' % (comm.getData( )))

  log.info("read remote IDS")
  comm = ReadRadioCtrlACL( serial=device.serial )
  device.execute(comm)
  log.info('comm:READ radio ACL: %r' % (comm.getData( )))

  log.info("read temporary basal")
  comm = ReadBasalTemp( serial=device.serial )
  device.execute(comm)
  log.info('comm:READ temp basal: %r' % (comm.getData( )))

  log.info("read settings")
  comm = ReadSettings( serial=device.serial )
  device.execute(comm)
  log.info('comm:READ settings!: %r' % (comm.getData( )))

  log.info("read contrast")
  comm = ReadContrast( serial=device.serial )
  device.execute(comm)
  log.info('comm:READ contrast: %r' % (comm.getData( )))

  log.info("read cur page number")
  comm = ReadCurPageNumber(serial=device.serial  )
  device.execute(comm)
  log.info('comm:READ page number!!!: %r' % (comm.getData( )))

  log.info("read HISTORY DATA")
  comm = ReadHistoryData(serial=device.serial, page=0)
  device.execute(comm)
  log.info('comm:READ history data page!!!:\n%s' % (comm.getData( )))

def get_pages(device):
  log.info("read cur page number")
  comm = ReadCurPageNumber(serial=device.serial )
  device.execute(comm)
  pages = comm.getData( )
  log.info('attempting to read %s pages of history' % pages)

  for x in range(pages + 1):
    log.info('comm:READ HISTORY DATA page number: %r' % (x))
    comm = ReadHistoryData(serial=device.serial, params=[ x ] )
    device.execute(comm)
    page = comm.getData( )
    log.info("XXX: READ HISTORY DATA!!:\n%s" % lib.hexdump(page))
    time.sleep(.100)

__all__ = [
  'BaseCommand', 'KeypadPush', 'PowerControl', 'PowerControlOff',
  'PumpCommand', 'PumpResume', 'PumpSuspend',
  'ReadBasalTemp', 'ReadBatteryStatus', 'ReadContrast',
  'ReadCurPageNumber', 'ReadErrorStatus', 'ReadFirmwareVersion',
  'ReadGlucoseHistory', 'ReadHistoryData', 'ReadPumpID',
  'ReadPumpModel', 'ReadPumpState', 'ReadPumpStatus',
  'ReadRTC', 'ReadRadioCtrlACL', 'ReadRemainingInsulin',
  'ReadRemainingInsulin523',
  'ReadSettings', 'ReadSettings523', 'ReadTotalsToday', 'SetSuspend',
  'PushEASY', 'PushUP', 'PushDOWN', 'PushACT', 'PushESC',
  'TempBasal', 'ManualCommand', 'ReadCurGlucosePageNumber',
  'SetAutoOff',
  'SetEnabledEasyBolus',
  'SetBasalType',
  'TempBasalPercent',
  'Bolus',
  'ReadErrorStatus508',
  'ReadBolusHistory',
  'ReadDailyTotals',
  'ReadPrimeBoluses',
  'ReadAlarms',
  'ReadProfileSets',
  'ReadUserEvents',
  'ReadRemoteControlID',
  'Read128KMem',
  'Read256KMem',
  'ReadBasalTemp508',
  'ReadTodayTotals508',
  'ReadSensorSettings',
  'ReadSensorHistoryData',
  'ReadISIGHistory',
  'FilterHistory',
  'FilterGlucoseHistory',
  'FilterISIGHistory',

  'ReadProfiles511_STD',
  'ReadProfiles511_A',
  'ReadProfiles511_B',
  'Model511_ExperimentOP125',
  'Model511_ExperimentOP126',
  'ReadSettings511',
  'ReadPumpTrace',
  'ReadDetailTrace',
  'Model511_Experiment_OP165',
  'ReadNewTraceAlarm',
  'ReadOldTraceAlarm',
  'WriteGlucoseHistoryTimestamp',
  'ReadLanguage',
  'ReadBolusWizardSetupStatus',
  'ReadCarbUnits',
  'ReadBGUnits',
  'ReadCarbRatios',
  'ReadCarbRatios512',
  'ReadInsulinSensitivities',
  'ReadBGTargets',
  'ReadBGTargets515',
  'ReadBGAlarmCLocks',
  'ReadReservoirWarning',
  'ReadBGReminderEnable',
  'ReadSettings512',
  'ReadProfile_STD512',
  'ReadProfile_A512',
  'ReadProfile_B512',
  'ReadLogicLinkIDS',
  'Model512Experiment_OP150',
  'ReadBGAlarmEnable',
  'GuardianSensorSettings',
  'GuardianSensorSettings',
  'GuardianSensorDemoGraphTimeout',
  'GuardianSensorAlarmSilence',
  'GuardianSensorRateChangeAlerts',
  'ReadSavedSettingsDate',
  'ReadBolusReminderEnable',
  'ReadBolusReminders',
  'ReadFactoryParameters',
  'ReadCalibrationFactor',
  'ReadVCNTRHistory',
  'ReadOtherDevicesIDS',
  'PumpTraceSelect',
  'PumpEnableDetailTrace',
  'PumpDisableDetailTrace',
  'Experiment_OP161',
  'Experiment_OP162',
  'Model511_Experiment_OP119',
  'Model511_Experiment_OP120',
  'Model511_Experiment_OP121',
  'Model511_Experiment_OP122',
  'Model511_Experiment_OP123',
  'Model511_Experiment_OP124',
  'Model511_Experiment_OP125',
  'Model511_Experiment_OP126',
  'Model511_Experiment_OP127',
  'Model511_Experiment_OP128',
  'Model511_Experiment_OP129',
  'Model511_Experiment_OP130',
  'SelectBasalProfile',
  'SelectBasalProfileSTD',
  'SelectBasalProfileA',
  'SelectBasalProfileB',
  'PumpExperiment_OP69',
  'PumpExperiment_OP70',
  'PumpExperiment_OP71',
  'PumpExperiment_OP72',
  'PumpExperiment_OP73',
  'PumpExperiment_OP75',
]

if __name__ == '__main__':
  import doctest
  doctest.testmod( )

  import sys
  port = None
  port = sys.argv[1:] and sys.argv[1] or False
  serial_num = sys.argv[2:] and sys.argv[2] or False
  if not port or not serial_num:
    print "usage:\n%s <port> <serial>, eg /dev/ttyUSB0 208850" % sys.argv[0]
    sys.exit(1)
  import link
  import stick
  import session
  from pprint import pformat
  logging.basicConfig(stream=sys.stdout, level=logging.DEBUG)
  log.info("howdy! I'm going to take a look at your pump and grab lots of info.")
  stick = stick.Stick(link.Link(port, timeout=.400))
  stick.open( )
  session = session.Pump(stick, serial_num)
  log.info(pformat(stick.interface_stats( )))
  log.info('PUMP MODEL: %s' % session.read_model( ))
  do_commands(session)
  log.info(pformat(stick.interface_stats( )))
  #get_pages(session)
  #log.info(pformat(stick.interface_stats( )))
  log.info("howdy! we downloaded a lot of pump info successfully.")
  # stick.open( )





import time
import logging
log = logging.getLogger( ).getChild(__name__)

import lib
import commands

class Downloader(object):
  log_format = 'logs/'
  def __init__(self, stick=None, device=None, log_format=log_format):
    self.stick = stick
    stick.open( )
    self.device = device
    self.log_format = log_format

  def download(self):
    """
    Download a single page, copy paste from elsewhere.
    """

    log.info("read HISTORY DATA")
    comm = commands.ReadHistoryData(serial=self.device.serial, page=0)
    self.device.execute(comm)
    log.info('comm:READ history data page!!!:\n%s' % (comm.getData( )))
    comm.save(prefix=self.log_format)

class PageDownloader(Downloader):
  log_format = 'logs/'
  def __init__(self, stick=None, device=None, log_format=log_format):
    self.stick = stick
    stick.open( )
    self.device = device
    self.log_format = log_format

  def read_current(self):

    log.info("read cur page number")
    comm = commands.ReadCurPageNumber( )
    self.device.execute(comm)
    self.pages = comm.getData( )
    log.info('attempting to read %s pages of history' % self.pages)
    return self.pages


  def download_page(self, x):
    log.info('comm:XXX:READ HISTORY DATA page number: %r' % (x))
    comm = commands.ReadHistoryData(serial=self.device.serial, params=[ x ] )
    self.device.execute(comm)
    page = comm.getData( )
    comm.save(prefix=self.log_format)
    log.info("XXX: READ HISTORY DATA page number %r!!:\n%s" % (x, page))
    time.sleep(.100)

  def download(self):
    self.read_current( )
    for x in range(self.pages + 1):
      log.info("read page %s" % x)
      self.download_page(x)

if __name__ == '__main__':
  import doctest
  doctest.testmod( )

  import sys
  port = None
  port = sys.argv[1:] and sys.argv[1] or False
  serial_num = sys.argv[2:] and sys.argv[2] or False
  if not port or not serial_num:
    print "usage:\n%s <port> <serial>, eg /dev/ttyUSB0 208850" % sys.argv[0]
    sys.exit(1)
  import link
  import stick
  import session
  from pprint import pformat
  logging.basicConfig(stream=sys.stdout, level=logging.DEBUG)
  log.info("howdy! I'm going to take a look at your pump download something info.")
  stick = stick.Stick(link.Link(port, timeout=.400))
  stick.open( )
  session = session.Pump(stick, serial_num)
  log.info(pformat(stick.interface_stats( )))

  downloader = Downloader(stick, session)
  downloader.download( )

  downloader = PageDownloader(stick, session)
  downloader.download( )

  log.info(pformat(stick.interface_stats( )))
  log.info("howdy! we downloaded everything.")



import time
from pprint import pformat
from cli import CommandApp
from decocare import link, stick, session, commands, lib, scan

def get_parser ( ):
  app = SendMsgApp( )
  return app.get_parser( )

class SendMsgApp(CommandApp):
  """
  %(prog)s - send messages to a compatible MM insulin pump

  This tool is intended to help discover protocol behavior.
  Under no circumstance is it intended to deliver therapy.
  """
  def main (self, args):
    if args.prefix:
      self.execute_list(args.prefix, args.saveall)
    if args.command == "ManualCommand":
      kwds = {
          'params': map(int, getattr(args, 'params', [ ])),
          'retries': getattr(args, 'retries', 0),
          'effectTime': getattr(args, 'effectTime'),
          'maxRecords': getattr(args, 'maxRecords'),
          'bytesPerRecord': getattr(args, 'bytesPerRecord'),
          'code': args.code,
          'name': getattr(args, 'name', "ExperimentCode%02x" % args.code),
          'descr': getattr(args, 'descr', "Experimental msg")
        }
      msg = commands.ManualCommand
      resp = self.exec_request(self.pump, msg, args=kwds,
                   dryrun=args.dryrun, render_hexdump=False, save=args.save, prefix=args.prefix_path)
    if args.command == "sleep":
      time.sleep(args.timeout)
    if args.command == "tweak":
      Other = getattr(commands, args.other)
      kwds = commands.TweakAnotherCommand.get_kwds(Other, args)
      print Other, kwds
      resp = self.exec_request(self.pump, Other, args=kwds,
                   dryrun=args.dryrun, save=args.save, prefix=args.prefix_path)
    if args.postfix:
      self.execute_list(args.postfix, args.saveall)

  def execute_list (self, messages, save=False):
    for name in messages:
      msg = getattr(commands, name)
      print "###### sending `%s`" % getattr(msg, 'name', msg)
      resp = self.exec_request(self.pump, msg, dryrun=self.args.dryrun,
                          render_hexdump=self.args.verbose>0,
                          save=save, prefix=self.args.prefix_path)

  def customize_parser (self, parser):
    choices = commands.__all__
    choices.sort( )
    parser.add_argument('--prefix-path',
                        dest="prefix_path",
                        type=str,
                        default="",
                        help="Prefix to store saved files when using --save or --saveall."
                        )
    parser.add_argument('--saveall',
                        action="store_true",
                        default=False,
                        help="Whether or not to save all responses.",
                        )

    parser.add_argument('--prefix',
                        action="append",
                        choices=choices,
                        help="Built-in commands to run before the main one."
                        )

    parser.add_argument('--postfix',
                        action="append",
                        choices=choices,
                        help="Built-in commands to run after the main one."
                        )

    subparsers = parser.add_subparsers(help="Main thing to do between --prefix and--postfix", dest="command")
    sleep_parser = subparsers.add_parser("sleep", help="Just sleep between command sets")
    sleep_parser.add_argument('timeout', type=float, default=0,
                              help="Sleep in between running --prefix and --postfix"
                              )
    tweaker = subparsers.add_parser("tweak", help="Tweak a builtin command")
    tweaker.add_argument('other',
                        choices=choices,
                        help="Command to tweak."
                        )
    commands.TweakAnotherCommand.config_argparse(tweaker)
    all_parser = subparsers.add_parser("ManualCommand", help="Customize a command")
    all_parser.add_argument('code', type=int,
                            help="The opcode to send to the pump."
                           )
    #__fields__ = ['maxRecords', 'code', 'descr',
    #            'serial', 'bytesPerRecord', 'retries', 'params']
    all_parser.add_argument('--params', type=str, action="append",
                            help="parameters to format into sent message",
                            default=commands.ManualCommand.params
                           )
    all_parser.add_argument('--params_hexline', dest='params', type=lib.decode_hexline,
                            help="hex string, parameters to format into sent message"
                            # default=commands.ManualCommand.params
                           )


    all_parser.add_argument('--descr', type=str, default="Experimental command",
                            help="Description of command"
                           )
    all_parser.add_argument('--name', type=str,
                            help="Proposed name of command"
                           )
    all_parser.add_argument('--save', action="store_true", default=False,
                            help="Save response in a file."
                           )
    all_parser.add_argument('--effectTime', type=float,
                            help="time to sleep before responding to message, float in seconds",
                            default=commands.ManualCommand.effectTime
                           )
    all_parser.add_argument('--maxRecords', type=int,
                            help="number of frames in a packet composing payload response",
                            default=commands.ManualCommand.maxRecords
                           )
    all_parser.add_argument('--bytesPerRecord', type=int,
                            help="bytes per frame",
                            default=commands.ManualCommand.bytesPerRecord
                           )

    return parser




#!/usr/bin/env python
# PYTHON_ARGCOMPLETE_OK

from decocare import commands
from decocare.helpers import cli
import argparse
import json

class TempBasalApp (cli.CommandApp):
  """ %(prog)s -  query or set temp basals

  Set or query temp basals.
  """
  def customize_parser (self, parser):
    parser.add_argument('command',
                        choices=['query', 'set', 'percent'],
                        default='query',
                        help="Set or query pump status [default: %(default)s)]"
                        )
    parser.add_argument('--duration',
                        dest='duration',
                        type=int, default=0,
                        help="Duration of temp rate [default: %(default)s)]"
                        )
    parser.add_argument('--rate',
                        dest='rate',
                        type=float, default=0,
                        help="Rate of temp basal [default: %(default)s)]"
                        )
    parser.add_argument('--out',
            dest="out",
            default='-',
            type=argparse.FileType('w'),
            help="Put basal in this file"
          )
    """
    subparsers = parser.add_subparsers(help="Main thing to do", dest="command")
    basals_parser = subparsers.add_parser("set", help="Just basals between command sets")
    basals_parser.add_argument('--duration',
                        dest='duration',
                        type=int, default=0,
                        help="Duration of temp rate [default: %(default)s)]"
                        )
    basals_parser.add_argument('--rate',
                        dest='rate',
                        type=float, default=0,
                        help="Rate of temp basal [default: %(default)s)]"
                        )
    """
    return parser
  def main (self, args):
    basals = self.query_temp(args)
    if args.command == "query":
      pass
    if args.command == "percent":
      params = format_percent_params(args)
      kwds = dict(params=params)
      msg = commands.TempBasalPercent
      resp = self.exec_request(self.pump, msg, args=kwds,
                   dryrun=args.dryrun, render_hexdump=True )
      basals = self.query_temp(args)
    if args.command == "set":
      params = format_params(args)
      kwds = dict(params=params)
      msg = commands.TempBasal
      resp = self.exec_request(self.pump, msg, args=kwds,
                   dryrun=args.dryrun, render_hexdump=True )
      basals = self.query_temp(args)

      # (serial=pump.serial, params=params)
    args.out.write(json.dumps(basals, indent=2))

  def query_temp (self, args):
    query = commands.ReadBasalTemp

    resp = self.exec_request(self.pump, query,
                 dryrun=args.dryrun, render_hexdump=False)
    results = resp.getData( )
    return results;

def format_percent_params (args):
  duration = int(args.duration / 30)
  rate = int(args.rate)
  params = [rate, duration]
  return params

def format_params (args):
  duration = args.duration / 30
  rate = int(args.rate / 0.025)
  params = [0x00, rate, duration]
  return params

if __name__ == '__main__':
  app = TempBasalApp( )
  app.run(None)




#!/usr/bin/env python
# PYTHON_ARGCOMPLETE_OK

from decocare import commands
from decocare.helpers import cli
import argparse
from decocare import lib
from dateutil.tz import gettz
from datetime import datetime
import time
import json

class SetRTCApp (cli.CommandApp):
  """ %(prog)s -  query or set RTC

  Set or query RTC.
  """
  def customize_parser (self, parser):
    parser.add_argument('command',
                        choices=['query', 'set', ],
                        default='query',
                        help="Set or query pump status [default: %(default)s)]"
                        )
    parser.add_argument('--rtc-out',
            dest="rtc_archive",
            default='-',
            type=argparse.FileType('w'),
            help="Put clock json in this file"
          )
    parser.add_argument('--timezone',
            default=gettz( ),
            type=gettz,
            help="Timezone to use"
          )
    parser.add_argument('--set',
            dest="set",
            required=True,
            type=lib.parse.date,
            help="Set clock to new value (iso8601)"
          )
    parser.add_argument('--out',
            dest="out",
            default='-',
            type=argparse.FileType('w'),
            help="Put basal in this file"
          )
    return parser
  def report_clock (self, args):
    self.clock = self.exec_request(self.pump, commands.ReadRTC)
    new_time = lib.parse.date(self.clock.getData( ))
    self.time = new_time
    return self.render_clock(self.clock)

  def render_clock (self, clock):
    new_time = lib.parse.date(clock.getData( ))
    new_time =  new_time.replace(tzinfo=self.args.timezone)
    results = dict(clock=new_time.isoformat( )
              , observed_at=datetime.now(self.args.timezone).isoformat( )
              , model=self.pump.model.model
              , _type='RTC')

    print "```json"
    self.args.rtc_archive.write(json.dumps(results, indent=2))
    print ''
    print "```"
    return results

  def set_clock (self, args):
    msg = commands.SetRTC
    kwds = dict(params=commands.SetRTC.fmt_datetime(args.set))
    new_clock = self.exec_request(self.pump, msg, args=kwds)
    # self.render_clock(new_clock)
    return new_clock
  def main (self, args):
    if args.dryrun:
      print args.set
      print lib.hexdump(bytearray(commands.SetRTC.fmt_datetime(args.set)))
      return
    clock = self.report_clock(args)
    if args.command == "query":
      pass
    if args.command == "set":

      new_clock = self.set_clock(args)

      time.sleep(1)
      clock = self.report_clock(args)

    args.out.write(json.dumps(clock, indent=2))

if __name__ == '__main__':
  app = SetRTCApp( )
  app.run(None)




#!/usr/bin/env python
# PYTHON_ARGCOMPLETE_OK

from decocare import commands
from decocare import lib
from decocare.helpers import cli

class BolusApp (cli.CommandApp):
  """ %(prog)s - Send bolus command to a pump.

  XXX: Be careful please!

  Units might be wrong.  Keep disconnected from pump until you trust it by
  observing the right amount first.
  """
  def customize_parser (self, parser):
    parser.add_argument('units',
                         type=float,
                         help="Amount of insulin to bolus."
                       )

    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument('--515',
                        dest='strokes_per_unit',
                        action='store_const',
                        const=10
                      )
    group.add_argument('--554',
                        dest='strokes_per_unit',
                        action='store_const',
                        const=40
                      )
    group.add_argument('--strokes',
                        dest='strokes_per_unit',
                        type=int
                      )

    return parser
  def main (self, args):
    print args
    self.bolus(args);

  def bolus (self, args):
    query = commands.Bolus
    kwds = dict(params=fmt_params(args))

    resp = self.exec_request(self.pump, query, args=kwds,
                 dryrun=args.dryrun, render_hexdump=False)
    return resp

def fmt_params (args):
  strokes = int(float(args.units) * args.strokes_per_unit)
  if (args.strokes_per_unit > 10):
    return [lib.HighByte(strokes), lib.LowByte(strokes)]
  return [strokes]

if __name__ == '__main__':
  app = BolusApp( )
  app.run(None)



import logging
import time

log = logging.getLogger( ).getChild(__name__)
import commands
import lib
import models
from errors import StickError, AckError, BadDeviceCommError


class Session(object):
  def __init__(self, stick):
    self.stick = stick
    self.should_download = True
  def init(self, skip_power_control=False):
    stick = self.stick
    log.info('test fetching product info %s' % stick)
    log.info(pformat(stick.product_info( )))
    log.info('get signal strength of %s' % stick)
    signal = 0
    while signal < 50:
      signal = stick.signal_strength( )
    log.info('we seem to have found a nice signal strength of: %s' % signal)
  def end(self):
    stick = self.stick
    log.info(pformat(stick.usb_stats( )))
    log.info(pformat(stick.radio_stats( )))

  def execute(self, command):
    self.command = command
    for i in xrange(max(1, self.command.retries)):
      log.info('execute attempt: %s' % (i + 1))
      try:
        self.expectedLength = self.command.bytesPerRecord * self.command.maxRecords
        self.transfer( )
        if self.should_download:
          log.info('sleeping %s before download' % command.effectTime)
          time.sleep(command.effectTime)
          self.download( )
        log.info('finished executing:%s' % command)
        if command.done( ):
          return command
      except BadDeviceCommError, e:
        log.critical("ERROR: %s" % e)
        # self.clearBuffers( )
    log.critical('this seems like a problem')

  def download(self):
    errors = [ ]
    if self.expectedLength > 0:
      log.info('proceeding with download')
      data = self.stick.download( )
      #self.command.data = data
      self.command.respond(data)
      return data
    else:
      log.info('no download required')
    assert not errors, ("with errors:%s" %"\n".join( map(str, errors) ))
  def transfer(self):
    log.info('session transferring packet')
    return self.stick.transmit_packet(self.command)

class Pump(Session):
  def __init__(self, stick, serial='208850'):
    super(type(self), self).__init__(stick)
    self.serial = serial
    log.info('setting up to talk with %s' % serial)

  def power_control(self, minutes=None):
    log.info('BEGIN POWER CONTROL %s' % self.serial)
    # print "PowerControl SERIAL", self.serial
    response = self.query(commands.PowerControl, minutes=minutes)
    power = self.command
    log.info('manually download PowerControl serial %s' % self.serial)
    data = self.stick.download( )
    log.info("ENDING manual download:\n%s" % lib.hexdump(data))
    return data

  def setModel (self, model=None, number=None):
    if number:
      self.model = models.lookup(number, self)
    if model:
      self.model = model
  def read_model(self):
    model = self.query(commands.ReadPumpModel)
    self.modelNumber = model.getData( )
    self.setModel(number=self.modelNumber)
    return model

  def read_history_0(self):
    log.info("read HISTORY DATA")
    comm = commands.ReadHistoryData(serial=self.serial, page=0)
    self.execute(comm)
    log.info('comm:READ history data page!!!:\n%s' % (comm.getData( )))

  def execute(self, command):
    command.serial = self.serial
    return super(type(self), self).execute(command)
  def query(self, Command, **kwds):
    command = Command(serial=self.serial, **kwds)
    self.execute(command)
    return command

if __name__ == '__main__':
  import doctest
  doctest.testmod( )

  import sys
  port = None
  port = sys.argv[1:] and sys.argv[1] or False
  serial_num = sys.argv[2:] and sys.argv[2] or False
  if not port or not serial_num:
    print "usage:\n%s <port> <serial>, eg /dev/ttyUSB0 208850" % sys.argv[0]
    sys.exit(1)
  import link
  import stick
  from pprint import pformat
  logging.basicConfig(stream=sys.stdout, level=logging.DEBUG)
  log.info("howdy! I'm going to take a look at your pump.")
  stick = stick.Stick(link.Link(port, timeout=.200))
  stick.open( )
  session = Pump(stick, serial_num)
  log.info(pformat(stick.interface_stats( )))
  session.power_control( )
  log.info(pformat(stick.interface_stats( )))
  log.info('PUMP MODEL: %s' % session.read_model( ))
  log.info(pformat(stick.interface_stats( )))
  log.info('PUMP READ PAGE 0: %s' % session.read_history_0( ))
  log.info(pformat(stick.interface_stats( )))
  log.info("howdy! all done looking at pump")
  # stick.open( )

#####
# EOF



#!/usr/bin/env python
# PYTHON_ARGCOMPLETE_OK

from decocare import commands
from decocare.helpers import cli

class PressKeysApp (cli.CommandApp):
  """%(prog)s - Simulate presses on the keypad.


  Press keys on the keypad.
  """
  def customize_parser (self, parser):
    parser.add_argument('commands',
                        nargs="+",
                        choices=['act', 'esc', 'up', 'down', 'easy' ],
                        # default='act',
                        help="buttons to press [default: %(default)s)]"
                        )
    return parser

  def exec_request (self, pump, msg, **kwds):
    msg = lookup_command(msg)
    super(PressKeysApp, self).exec_request(pump, msg, **kwds)

command_map = {
    'easy': commands.PushEASY,
    'esc': commands.PushESC,
    'act': commands.PushACT,
    'down': commands.PushDOWN,
    'up': commands.PushUP
}

def lookup_command (name):
  return command_map.get(name)

if __name__ == '__main__':
  app = PressKeysApp( )
  app.run(None)





from datetime import datetime

class NotADate(Exception): pass

class Mask:
  """
  Some useful bit masks.
  """
  time   = 0xC0
  invert = 0x3F
  year   = 0x7F
  day    = 0x1F


def quick_hex(bb):
  return ' '.join( [ '%#04x' % x for x in bb ] )

def parse_seconds(sec):
  """
  simple mask
  >>> parse_seconds(0x92)
  18
  """
  return sec & Mask.invert

def parse_minutes(minutes):
  """
  simple mask
  >>> parse_minutes(0x9e)
  30
  """
  return minutes & Mask.invert


def parse_hours(hours):
  """
  simple mask
  >>> parse_hours(0x0b)
  11

  >>> parse_hours(0x28)
  8

  """
  return int(hours & 0x1f)

def parse_day(day):
  """
  simple mask
  >>> parse_day( 0x01 )
  1
  """
  return day & Mask.day

def parse_months(seconds, minutes):
  """
  calculate from extra bits shaved off of seconds and minutes:

  >>> parse_months( 0x92, 0x9e )
  10

  """
  high = (seconds & Mask.time) >> 4
  low  = (minutes & Mask.time) >> 6
  return high | low

def parse_years_lax(year):
  """
  simple mask plus correction
  """
  y = (year & Mask.year) + 2000
  return y


def extra_year_bits(year=0x86):
  """
  practice getting some extra bits out of the year byte
  >>> extra_year_bits( )
  [1]

  >>> extra_year_bits(0x06)
  [0]

  >>> extra_year_bits(0x86)
  [1]

  >>> extra_year_bits(0x46)
  [0]

  >>> extra_year_bits(0x26)
  [0]

  >>> extra_year_bits(0x16)
  [0]

  """
  # year = year[0]
  masks = [ ( 0x80, 7) ]
  nibbles = [ ]
  for mask, shift in masks:
    nibbles.append( ( (year & mask) >> shift ) )
  return nibbles

def extra_hour_bits(value):
  """
  practice getting extra bits out of the hours bytes

  >>> extra_hour_bits(0x28)
  [0, 0, 1]

  >>> extra_hour_bits(0x8)
  [0, 0, 0]
  """
  masks = [ ( 0x80, 7), (0x40, 6), (0x20, 5), ]
  nibbles = [ ]
  for mask, shift in masks:
    nibbles.append( ( (value & mask) >> shift ) )
  return nibbles

def parse_years(year):
  """
    simple mask plus correction
    >>> parse_years(0x06)
    2006

  """
  # if year > 0x80:
  #  year = year - 0x80
  y = (year & Mask.year) + 2000
  # if y < 0 or y < 1999 or y > 2015:
  #   raise ValueError(y)
  return y

def encode_year(year):
  pass

def encode_monthbyte(sec=18, minute=30, month=10):
  """
  >>> encode_monthbyte( ) == bytearray(b'\x92\x9e')
  True

  >>> quick_hex(encode_monthbyte( ))
  '0x92 0x9e'

  >>> encode_monthbyte(sec=10) == bytearray(b'\x8a\x9e')
  True

  >>> encode_monthbyte(sec=35) == bytearray(b'\xa3\x9e')
  True

  >>> encode_monthbyte(sec=50) == bytearray(b'\xb2\x9e')
  True

  >>> encode_monthbyte(minute=10) == bytearray(b'\x92\x8a')
  True

  >>> encode_monthbyte(minute=35) == bytearray(b'\x92\xa3')
  True

  >>> encode_monthbyte(minute=50) == bytearray(b'\x92\xb2')
  True

  """

  encoded = [ 0x00, 0x00 ]


  high = (month & (0x3 << 2)) >> 2
  low  = month & (0x3)

  encoded[0] = sec | (high << 6)
  encoded[1] = minute | (low << 6)
  return bytearray( encoded )

  printf("0x%.2x 0x%.2x\n", buf[0], buf[1]);

def encode_minute(minute=30, month=10):
  """
  >>> quick_hex(encode_minute( ))
  '0x9e'

  """
  low  = month & (0x3)
  encoded = minute | (low << 6)
  return bytearray( [ encoded ] )

def encode_second(sec=18, month=10):
  """
  >>> encode_second( ) == bytearray(b'\x92')
  True

  >>> quick_hex(encode_second( ))
  '0x92'

  """
  high = (month & (0x3 << 2)) >> 2
  encoded = sec | (high << 6)
  return bytearray( [ encoded ] )

def test_time_encoders( ):
  """
  >>> test_time_encoders( )
  True
  """

  one = bytearray().join([encode_second( ), encode_minute( )])
  two = encode_monthbyte( )
  return one == two

def unmask_date(data):
  """
  Extract date values from a series of bytes.
  Always returns tuple given a bytearray of at least 5 bytes.

  Returns 6-tuple of scalar values year, month, day, hours, minutes,
  seconds.
  >>> unmask_date(bytearray( [ 0x6f, 0xd7, 0x08, 0x01, 0x06 ] ))
  (2006, 7, 1, 8, 23, 47)

  >>> unmask_date( bytearray( [ 0x00 ] * 5 ))
  (2000, 0, 0, 0, 0, 0)

  """
  data = data[:]
  seconds = parse_seconds(data[0])
  minutes = parse_minutes(data[1])

  hours   = parse_hours(data[2])
  day     = parse_day(data[3])
  year    = parse_years(data[4])

  month   = parse_months( data[0], data[1] )
  return (year, month, day, hours, minutes, seconds)

def parse_date(date, strict=False, loose=False):
  """
  Choose whether or not to throw an error if you get a bad date.
  """
  try:
    return parse_date_strict(date)
  except NotADate, e:
    if strict:
      raise
    if loose:
      return unmask_date(data)

  return None

def parse_date_strict(data):
  """
  Apply strict datetime validation to extract a datetime from a series
  of bytes.
  Always throws an error for bad dates.

  >>> parse_date(bytearray( [ 0x6f, 0xd7, 0x08, 0x01, 0x06 ] )).isoformat( )
  '2006-07-01T08:23:47'


  """
  (year, month, day, hours, minutes, seconds) = unmask_date(data)
  try:
    date = datetime(year, month, day, hours, minutes, seconds)
    return date
  except ValueError, e:
    raise NotADate(e)

__test__ = {
  'unmask': '''
  These examples are fixed! Thanks ehwest, robg.
  >>> unmask_date( bytearray( [ 0x93, 0xd4, 0x0e, 0x10, 0x0c ] ))
  (2012, 11, 16, 14, 20, 19)

  >>> unmask_date( bytearray( [ 0xa6, 0xeb, 0x0b, 0x10, 0x0c, ] ))
  (2012, 11, 16, 11, 43, 38)

  >>> unmask_date( bytearray( [ 0x95, 0xe8, 0x0e, 0x10, 0x0c ] ))
  (2012, 11, 16, 14, 40, 21)


  >>> unmask_date( bytearray( [ 0x80, 0xcf, 0x30, 0x10, 0x0c, ] ))
  (2012, 11, 16, 16, 15, 0)


  >>> unmask_date( bytearray( [ 0xa3, 0xcf, 0x30, 0x10, 0x0c, ] ))
  (2012, 11, 16, 16, 15, 35)

''',

  'encode_monthbyte': '''
  >>> encode_monthbyte(month=1) == bytearray(b'\x12^')
  True

  >>> encode_monthbyte(month=2) == bytearray(b'\x12\x9e')
  True

  >>> encode_monthbyte(month=3) ==  bytearray(b'\x12\xde')
  True

  >>> encode_monthbyte(month=10, minute=0, sec=0) == bytearray(b'\x80\x80')
  True

  >>> encode_monthbyte(month=10, minute=0, sec=24) == bytearray(b'\x98\x80')
  True
''',

}

if __name__ == '__main__':
  import doctest
  doctest.testmod( )




import lib
import logging
import time

"""
stick - implement a naive open source driver for Medtronic's
Carelink USB stick.

Please ask Medtronic for additional information on how to use the usb
stick.

Consumes a :ref:`link`, which allows us to debug everything on the
wire.

This module contains a class for each logical operation available in
the stick's firmware.  Each operation is sent has a string of bytes
over the serial/usb interface, and then a response with a particular
format can be read.

"""

log = logging.getLogger( ).getChild(__name__)

from errors import StickError, AckError, BadDeviceCommError

class BadCRC(StickError): pass
class UnresponsiveError (StickError): pass

def CRC8(data):
  return lib.CRC8.compute(data)

class StickCommand(object):
  """Basic stick command

  Each command is used to talk to the usb stick.
  The usb stick interprets the opcode, and then performs the function
  associated with the opcode.
  Altogether, the suite of opcodes that the stick responds to allows
  you to debug and track all packets you are sending/receiving plus
  allows you to send recieve commands to the pump, by formatting your
  message into payloads with opcodes, and then letting the stick work
  on what you've given it.  It's kind of like a modem with this funky
  binary interface and 64 byte payloads.
  """
  code = [ 0x00 ]
  label = 'example stick command'
  delay = .001
  size = 64

  def __str__(self):
    code = ' '.join([ '%#04x' % op for op in self.code ])
    return '{0}:{1}'.format(self.__class__.__name__, code)
  def __repr__(self):
    return '<{0:s}:size({1})>'.format(self, self.size)
  def format(self):
    return self.format_cl2(*self.code)

  def format_cl2(self, msg, a2=0x00, a3=0x00):
    # generally commands are 3 bytes, most often CMD, 0x00, 0x00
    msg = bytearray([ msg, a2, a3 ])
    return msg

  def parse(self, data):
    self.data = data

  def respond(self, raw):
    if len(raw) == 0:
      log.error("ACK is zero bytes!")
      # return False
      raise AckError("ACK is 0 bytes:\n%s" % lib.hexdump(raw))
    commStatus = raw[0]
    # usable response
    assert commStatus == 1, ('commStatus: %02x expected 0x1' % commStatus)
    status     = raw[1]
    # status == 102 'f' NAK, look up NAK
    if status == 85: # 'U'
      return raw[:3], raw[3:]
    assert False, ("NAK!!\n%s" % lib.hexdump(raw[:3]))
    

class ProductInfo(StickCommand):
  """Get product info from the usb device.

  Useful for identifying
  what kind of usb stick you've got; there are a few different kinds.
  Eg, European vs US regulatory domains require different frequencies for compliance.

  """
  code   = [ 4 ]
  SW_VER = 16
  label  = 'usb.productInfo'
  rf_table  = { 001: '868.35Mhz' ,
                000: '916.5Mhz'  ,
                255: '916.5Mhz'  }
  iface_key = { 3: 'USB',
                1: 'Paradigm RF' }

  @classmethod
  def decodeInterfaces( klass, L ):
    n, tail    = L[ 0 ], L[ 1: ]
    interfaces = [ ]
    for x in xrange( n ):
      i    = x*2
      k, v = tail[i], tail[i+1]
      interfaces.append( ( k, klass.iface_key.get( v, 'UNKNOWN'  ) ) )
    return interfaces

  @classmethod
  def decode( klass,  data ):
    return {
      'rf.freq'          : klass.rf_table.get( data[ 5 ], 'UNKNOWN' )
    , 'serial'           : str( data[ 0:3 ]).encode( 'hex'  )
    , 'product.version'  : '{0}.{1}'.format( *data[ 3:5 ] )
    , 'description'      : str( data[ 06:16 ] )
    , 'software.version' : '{0}.{1}'.format( *data[ 16:18 ] )
    , 'interfaces'       : klass.decodeInterfaces( data[ 18: ] )
    }

  _test_ok = bytearray( [
  ] )
  def parse(self, data):
    """
      #>>>
    """
    return self.decode(data)

class InterfaceStats(StickCommand):
  """Abstract stats decoder.
  """
  code          = [ 5 ]
  INTERFACE_IDX = 19
  label         = 'usb.interfaceStats'
  @classmethod
  def decode( klass, data):
    """
    Decode interface stats.  The stick exposes 6 counters to monitor errors,
    crcs, naks, timeouts, rx, and tx packets.  Very useful for debugging.
    """
    return {
      'errors.crc'      : data[ 0 ]
    , 'errors.sequence' : data[ 1 ]
    , 'errors.naks'     : data[ 2 ]
    , 'errors.timeouts' : data[ 3 ]
    , 'packets.received': lib.BangLong( data[ 4: 8 ] )
    , 'packets.transmit': lib.BangLong( data[ 8:12 ] )
    }
  def parse(self, data):
    """
      #>>>
    """
    return self.decode(data)


class UsbStats(InterfaceStats):
  """Count of packets and stats on the usb side of the stick."""
  code = [ 5, 1 ]

class RadioStats(InterfaceStats):
  """Count of packets and stats on the radio side of the stick."""
  code = [ 5, 0 ]

class SignalStrength(StickCommand):
  """This seems to be required to initialize communications with the
  usb stick.  Also, you should wait until a minimum threshold is
  reached.
  """
  code = [ 6, 0 ]
  def parse(self, data):
    """
      #>>>
    """
    # result[0] is signal strength
    self.value = int(data[0])
    log.info('%r:readSignalStrength:%s' % (self, int(data[0])))
    return int(data[0])

class LinkStatus(StickCommand):
  """Basic ACK type of command.
  Used to poll the modem's radio buffer.  When the radio buffer is
  full, we can download a packet from the buffer.  Otherwise, we need
  to be mindful of the state the radio is in.  This opcode tells you
  the current state of the radio/stick.
  """
  code = [ 0x03 ]
  reasons = ['OK']

  def __str__(self):
    extra = ''
    size = getattr(self, 'size', None) or '??'
    extra += "size=%s" % size
    if getattr(self, 'error', False):
      extra += '{0}:error:{1}:reason:{2}'.format(self.__class__.__name__, self.error, str(self.reasons))
    base = super(type(self), self).__str__( )
    return '{0}:status:{1}'.format(base, extra)
      
  def record_error(self, result):
    self.error  = True
    self.ack    = result[0] # 0 indicates success
    # believe success = result[1] # 'U' or 'f'
    self.status = result[2]
    lb, hb      = result[3], result[4]
    self.size   = lib.BangInt((lb, hb))

    if self.ack == 0 and (self.status & 0x1) > 0:
      self.error = False
    self.set_reason(self.ack)

  def set_reason(self, status):
    reasons = [ ]
    if (status & 0x2) > 0:
      reasons.append('STATUS: receive in progress!')
    if (status & 0x4) > 0:
      reasons.append('STATUS: transmit in progress!')
    if (status & 0x8) > 0:
      reasons.append('STATUS: interface error!')
    if (status & 0x10) > 0:
      reasons.append('STATUS: receive overflow!')
    if (status & 0x20) > 0:
      reasons.append('STATUS: transmit overflow!')
    if (status & 0x1) > 0:
      reasons.append('STATUS: OK')
    msg = '\n'.join(map(str, [ self, '|'.join(reasons) ]))
    log.info(msg)
    self.reasons = reasons

  def parse(self, result):
    """
      #>>>
    """
    self.record_error(result)
    if self.ack != 0:
      log.error("readStatus: non-zero status: %02x" % self.ack)
      # should this trigger a retry, and if so where?
      # should the other usb commands also trigger these retries? and at the
      # same points?
      raise AckError("readStatus: non-zero status: %02x" % self.ack)
    if self.error is not True:
      return self.size
    return 0

class ReadRadio(StickCommand):
  """
  Read buffer from the radio.

  Downloads a packet from the radio buffer.

  """
  code = [ 0x0C, 0x00 ]
  dl_size = 0
  size = 64
  def __init__(self, size):
    self.size = size
    self.dl_size = size
    packet = [12, 0, lib.HighByte(size), lib.LowByte(size)]
    if size < 64 and size != 15:
      log.error('size (%s) is less than 64 and not 15, which may cause an error.' % size)
      self.size = 64
    self.code = packet + [ CRC8(packet) ]

  def __str__(self):
    return '{0}:size:{1}'.format(self.__class__.__name__, self.dl_size)
  def __repr__(self):
    return '<{0:s}>'.format(self)
  def format(self):
    msg = bytearray(self.code)
    return msg

  def respond(self, raw):
    if len(raw) == 0:
      log.error("ReadRadio ACK is zero bytes!")
      # return False
      raise AckError("ACK is 0 bytes: %s" % lib.hexdump(raw))
    log.info('readData validating remote raw[ack]: %02x' % raw[0])
    log.info('readData; foreign raw should be at least 14 bytes? %s %s' % (len(raw), len(raw) > 14))
    log.info('readData; raw[retries] %s' % int(raw[3]))
    dl_status = int(raw[0])
    if dl_status != 0x02: # this differs from the others?
      raise BadDeviceCommError("bad dl raw! %r" % raw)
      assert (int(raw[0]) == 2), repr(raw)
    return raw[:1], raw
    
  def parse(self, raw):
    """
    Detect BadCRC here.  Also, look for eod set.
    """
    """
    log.info('readData validating remote raw[ack]: %02x' % raw[0])
    log.info('readData; foreign raw should be at least 14 bytes? %s %s' % (len(raw), len(raw) > 14))
    log.info('readData; raw[retries] %s' % int(raw[3]))
    dl_status = int(raw[0])
    if dl_status != 0x02: # this differs from the others?
      raise BadDeviceCommError("bad dl raw! %r" % raw)
      assert (int(raw[0]) == 2), repr(raw)
    # raw[1] != 0 # interface number !=0
    # raw[2] == 5 # timeout occurred
    # raw[2] == 2 # NAK
    # raw[2] # should be within 0..4
    log.info("readData ACK")
    lb, hb    = raw[5] & 0x7F, raw[6]
    self.eod  = (raw[5] & 0x80) > 0
    """
    lb, hb    = raw[5] & 0x7F, raw[6]
    self.eod  = (raw[5] & 0x80) > 0
    resLength = lib.BangInt((lb, hb))
    # we don't really care about the length
    #assert resLength < 64, ("cmd low byte count:\n%s" % lib.hexdump(raw))

    data = raw[13:13+resLength]
    self.packet = data
    log.info('%s:eod:found eod (%s)' % (self, self.eod))
    log.info('found packet len(%s), link expects(%s)' % (len(self.packet), resLength))
    assert len(data) == resLength
    head = raw[13:]
    crc = raw[-1]
    # crc check
    if crc == 0 and len(data) > 1:
      log.warn('bad zero CRC?')
    expected_crc = CRC8(data)
    if crc != expected_crc:
      msg = ':'.join( [ 'ReadRadio:BAD ACK:found raw[crc]: %#04x' % (crc),
                          'expected_crc(data): %#04x' % (expected_crc),
                          'raw:\n%s\n' % (lib.hexdump(raw)),
                          'head:\n%s\n' % (lib.hexdump(head)),
                          'data:\n%s\n' % (lib.hexdump(data)) ] )
      log.info(msg)
      log.info("XXX:IGNORE:BadCRC:returning empty message, sleep .100, avoid errors.")
      time.sleep(.100)
      return bytearray( )
      raise BadCRC(msg)
    assert crc == expected_crc
    return data

class TransmitPacket(StickCommand):
  """Format a packet to send on the radio.

  This commands formats a packet from usb, and shoves it into the
  radio buffer.
  The radio buffer is broadcast "over the air" so that any device
  sensitive to the packets you sent will respond accordingly
  (probably sending data back).
  For this reason, the serial number of the device you'd like to talk
  to is formatted into the packet.

  """
  code = [ 1, 0, 167, 1 ]
  head = [ 1, 0, 167, 1 ]
  # wraps pump commands
  def __init__(self, command):
    self.command = command
    self.params  = command.params
    self.code    = command.code
    self.retries = command.retries
    self.serial  = command.serial
    # self.delay = command.effectTime

  def __str__(self):
    if getattr(self, 'command', False):
      return '{0}:{1:s}'.format(self.__class__.__name__, self.command)
    code = ' '.join([ '%#04x' % op for op in self.head ])
    return '{0}:{1}'.format(self.__class__.__name__, code)
  def __repr__(self):
    return '<{0:s}>'.format(self)
  

  def calcRecordsRequired(self):
    return self.command.calcRecordsRequired( )

  def format(self):
    """
    Formatting of the packet to be sent gets done here.
    """
    params = self.params
    code   = self.code
    maxRetries = self.retries
    serial = list(bytearray(self.serial.decode('hex')))
    paramsCount = len(params)
    head   = [ 1, 0, 167, 1 ]
    # serial
    packet = head + serial
    # paramCount 2 bytes
    packet.extend( [ (0x80 | lib.HighByte(paramsCount)),
                             lib.LowByte(paramsCount) ] )
    # not sure what this byte means
    button = 0
    # special case command 93
    if code == 93:
      button = 85
    packet.append(button)
    packet.append(maxRetries)
    # how many packets/frames/pages/flows will this take?
    responseSize = self.calcRecordsRequired()
    # really only 1 or 2?
    pages = responseSize
    if responseSize > 1:
      pages = 2
    packet.append(pages)
    packet.append(0)
    # command code goes here
    packet.append(code)
    packet.append(CRC8(packet))
    packet.extend(params)
    packet.append(CRC8(params))
    log.debug(packet)
    return bytearray(packet)

  def respond(self, raw):
    code = self.command.code
    params = self.params
    if code != 93 or params[0] != 0:
      ack, body = super(type(self), self).respond(raw)
      return ack, body
      
    return (bytearray(raw), bytearray(raw))
  def parse(self, results):
    return results
    #self.checkAck(results)



class Stick(object):
  """
  The carelink usb stick acts like a buffer.

  It has a variety of commands providing synchronous IO, eg, you may
  generally perform a read immediately after writing to it, and expect a
  response.

  The commands operate on a local buffer used to facilitate exchanging
  messages over RF with the pump. RF communication with the pump
  happens asynchronously, requiring us to go through 3 separate
  phases for each message we'd like to exchange with the pumps:
  * transmit - send commmand
  * poll_size - loop
  * download - loop

  Each command is usually only 3 bytes.

  The protocol offers some facility for detecting and recovering
  from inconsistencies in the underlying transport of data, however,
  we are somwhat ignorant of them.  The tricky bits are exactly how
  to recover from, eg CRC, errors that can occur.
  The "shape" and timing of these loops seem to mostly get the job
  done.

  The Stick object provides a bunch of useful methods, that given a link,
  will represent the state of one active usb stick.

  """
  link = None
  def __init__(self, link):
    self.link = link
    self.command = None
    self._download_i = False

  def __str__(self):
    s = [ self.__class__.__name__,
          'transmit[{}]'  .format(str(getattr(self, 'transmit', None))),
          'reader[{}]'    .format(str(getattr(self, 'reader', None))),
          'download_i[{}]'.format(str(getattr(self, '_download_i', None))),
          'status[{}]'    .format(repr(getattr(self, 'last_status', None))),
          'poll_size[{}]' .format(str(getattr(self, '_poll_size', None))),
          'poll_i[{}]'    .format(str(getattr(self, '_poll_i', None))),
          'command[{}]'   .format(repr(getattr(self, 'command', None))),
        ]
    return ' '.join(s)
  def __repr__(self):
    return '<{0}>'.format(str(self))
  def process(self):
    """
    Working with the usb stick typically follows a pretty routine process:
    1.  send our opcode, get a response
    2.  use some custom logic, per opcode to respond to the stick's reponse
    3.  parse the response from that, return result

    This has to be done for each opcode.
    """
    msg = ':'.join(['PROCESS', 'START'
           ] + map(str, [ self.timer.millis( ), self.command]))
    log.info(msg)
    log.info('link %s processing %s)' % ( self, self.command ))
    """
    self.link.write(self.command.format( ))
    log.debug('sleeping %s' % self.command.delay)
    time.sleep(self.command.delay)
    size = max(64, self.command.size)
    raw = bytearray(self.link.read(size))

    """
    raw = self.send_force_read( )
    if not raw or len(raw) == 0:
      log.info('process zero length READ, try once more sleep .010')
      time.sleep(.010)
      raw = bytearray(self.link.read(self.command.size))

    ack, response = self.command.respond(raw)
    info = self.command.parse(response)
    log.info('finished processing {0}, {1}'.format(self.command, repr(info)))
    msg = ':'.join(['PROCESS', 'END'
           ] + map(str, [ self.timer.millis( ), self.command]))
    log.info(msg)
    return info
    
  def query(self, Command):
    """
    query - simplify the process of working with the stick, pass your command, get the result
    """
    self.command = command = Command( )
    return self.process( )

  def product_info(self):
    """
    Get the product info from the connected stick.
    """
    return self.query(ProductInfo)

  def interface_stats(self):
    """
    debug both sets of interface stats.
    """
    return {'usb': self.usb_stats( ), 'radio': self.radio_stats( ) }
    
  def usb_stats(self):
    """
      just get usb stats.
    """
    return self.query(UsbStats)

  def radio_stats(self):
    """
      just get radio stats.
    """
    return self.query(RadioStats)

  def signal_strength(self):
    """
      just get signal strength from connected stick
    """
    return self.query(SignalStrength)

  def poll_size(self):
    """
      query how many bytes are waiting in the radio buffer, ready to be downloaded

      There seem to be a few sweet spots, where you want to download the data.
    """
    size  = 0
    start = time.time()
    i     = 0
    log.debug('%r:STARTING POLL PHASE:attempt:%s' % (self, i))
    #while size == 0 and size < 64 and time.time() - start < 1:
    while size == 0 and time.time() - start < 1:
      self._poll_i = i
      self._poll_size = size
      log.debug('%r:poll:attempt:%s' % (self, i))
      size  = self.read_status( )
      self._poll_size = size
      if size == 0:
        log.debug('poll zero, sleeping in POLL, .100')
        time.sleep(.100)
      i += 1
    log.info('%s:STOP POLL after %s attempts:size:%s' % (self, i, size))
    self._poll_size = size
    self._poll_i = False
    return size
    
  def read_status(self):
    """
    Get current link status.
    """
    # log.debug('read_status')
    result = self.query(LinkStatus)
    self.last_status = self.command
    return result

  def old_download_packet(self, size):
    """
    Naive version of downloading a packet.
    Didn't quite work right.
    """
    log.info("download_packet:%s" % (size))
    self.command = ReadRadio(size)
    packet = self.process( )
    return packet

  def send_force_read(self, retries=1, timeout=1):
    """
    Pretty simple, try really hard to ensure that we've sent our bytes, and we
    get a response.
    This is probably overkill, but seems to get the job done.
    """
    # 
    # so the behavior of a read_radio should probably be similar to
    # poll_size??
    reader = self.command
    read_size = 64
    size = reader.size
    start = time.time( )
    raw = bytearray( )
    for attempt in xrange(retries):
      log.info(' '.join([
        'send_force_read: attempt {0}/{1}'.format(attempt, retries),
        'send command,',
        'read until we get something within some timeout']))
      log.info('link %s sending %s)' % ( self, reader ))
      self.link.write(reader.format( ))
      log.debug('sleeping %s' % reader.delay)
      time.sleep(reader.delay)
      raw = bytearray(self.link.read(size))
      if len(raw) == 0:
        log.info('zero length READ, try once more sleep .250')
        time.sleep(.250)
        raw = bytearray(self.link.read(self.command.size))

      if len(raw) != 0:
        log.info(' '.join(['quit send_force_read,',
                           'found len:', str(len(raw)),
                           'expected', str(size),
                           'after', str(attempt), 'attempts']))
        return raw
    log.critical(' '.join([ "FAILED TO DOWNLOAD ANYTHING,",
                            "after %s " % (attempt),
                            "expected:%s" % (size) ]))
    assert not raw
    
  def download_packet(self, size):
    """
    This is the tricky bit, where we stroke the radio and hope it gives us a
    buffer full of data.
    """
    log.info("%s:download_packet:%s" % (self, size))
    # XXX: this is the tricky bit
    original_size = size
    self.command = reader = ReadRadio(size)
    self.reader = reader
    msg = ':'.join(['PROCESS', 'START'
           ] + map(str, [ self.timer.millis( ), self.command]))
    log.info(msg)
    if size == 0:
      log.info('Download Size is ZERO, returning nothing')
      return bytearray( )
      
    raw = self.send_force_read( )
    # return
    # packet = self.process( )
    # return packet

    # copy pasted from process
    """
    log.info('link %s processing %s)' % ( self, self.command ))
    # self.link.process(command)
    self.link.write(self.command.format( ))
    log.debug('sleeping %s' % self.command.delay)
    time.sleep(self.command.delay)
    size = max(64, self.command.size)
    raw = bytearray(self.link.read(size))
    """

    # if len(raw) == 0:
    if not raw:
      log.info('zero length READ, try once more sleep .500')
      time.sleep(.500)
      raw = bytearray(self.link.read(self.command.size))

    try:
      ack, response = self.command.respond(raw)
      info = self.command.parse(response)
      msg = ':'.join(['PROCESS', 'END'
             ] + map(str, [ self.timer.millis( ), self.command]))
      log.info(msg)
      return info
    except BadDeviceCommError, e:
      log.critical("download_packet:%s:ERROR:%s:ACK!?" % (self, e))
      log.info("we failed to pass %s ACK!?" % (self.command))
      log.info('expected size was: %s' % original_size)
      status = LinkStatus( )
      if original_size < 64:
        #size = self.read_status( )
        #size = self.poll_size( )
        log.info('XXX:JUST a bit more READ new size: %s, sleep .100' % original_size)
        self.link.write(status.format( ))
        time.sleep(.100)
        raw = bytearray(self.link.read(64))
        ack, response = reader.respond(raw)
        info = reader.parse(response)
        return info

      ack, body = status.respond(raw)
      size = status.parse(body)
      log.info('attempt another read')
      info = None

      raw = bytearray(self.link.read(size))
      if len(raw) == 0:
        log.info('NESTED zero length READ, try once more sleep .100')
        time.sleep(.100)
        raw = bytearray(self.link.read(self.command.size))

        ack, body = status.respond(raw)
        info = self.command.parse(body)



    log.info('finished processing {0}, {1}'.format(self.command, repr(info)))
    return info
    
  def download(self, size=None):
    """
    Theory is to download anything and everything available off the radio
    buffer, and to wait if necessary.
    """
    eod = False
    results = bytearray( )
    ailing = 0
    i = 0
    log_head = 'download(attempts[{}])'
    expecting = 'download(attempts[{}],expect[{}])'
    stats = '{}:download(attempts[{}],expect[{}],results[{}]:data[{}])'
    expect_eod = False
    log.info('download:start:%s' % i)
    data = bytearray( )
    while not eod:
      i += 1
      self._download_i = i
      data = bytearray( )
      if size is None:
        log.info("%s:begin first poll first sleep .250" % (stats.format(self, i, 0,
                                          len(results), len(data))))
        time.sleep(.250)
        size = self.poll_size( )
        log.info("%s:end first poll" % (stats.format(self, i, size,
                                        len(results), len(data))))
      if size == 0:
        if i % 3 == 0:
          time.sleep(1.5)
        #time.sleep(1.5)
        size = self.poll_size( )
      """
      if size == 0:
      # if size == 0 and i > 1:
        log.info("%s:zero poll size, sleep .500 try again" % ( \
                  stats.format(self, i, size, len(results), len(data))))
        size = self.poll_size( )
        time.sleep(.500)
      """
      if size == 0 and i > 1:
        log.warn("%s:BAD AILING" % (stats.format(self, i, size,
                                        len(results), len(data))))
        ailing = ailing + 1
        if ailing > 1:
          break
        continue
          # break
      elif ailing > 0:
        ailing = ailing - 1

      log.info("%s:proceed to download packet" % (stats.format(self, i, size,
                                                  len(results), len(data))))
      #time.sleep(.100)
      data = self.download_packet(size)
      expect_eod = False
      if data:
        results.extend(data)
        expect_eod = self.command.eod 
        log.info("%s:adding segment" % (stats.format(self, i, size,
                                        len(results), len(data))))
      else:
        log.info("%s:no data, try again sleep .400" % (stats.format(self, i, size,
                                            len(results), len(data))))
        time.sleep(.400)
      # eod = expect_eod and size < 15
      eod = expect_eod
      # or size < 15
      if not eod:
        log.info("%s:no eod, sleep .200 try again" % (stats.format(self, i, size,
                                            len(results), len(data))))
        time.sleep(.200)
        size = self.poll_size( )

    log.info("%s:DONE" % (stats.format(self, i, size,
                          len(results), len(data))))
    self._download_i = False
    # self.reader = None
    return results

  def clear_buffer(self):
    """
    An alternative download solution.  This can be helpful in
    scenarios where a prior run seems crashed your process, but the
    radio is still transmitting and receiving data.  Running this
    method collects data from the radio until it's done receiving,
    more or less, at which point you should be free to try again.
    """
    bad = bytearray( )
    raw = bytearray( )
    for attempt in xrange( 3 ):
      segments = [ ]
      segs_vs_raw = 'segments[{0}],total_segments[{1}]:raw[{2}]'
      seg_stats = ( len(segments), sum(map(len, segments)), len(raw) )
      log_detail = segs_vs_raw.format(*seg_stats)
      log_head = "XXX:clear_buffer[attempt][%s]" % (attempt)
      log.debug('INTERFACE STATS:\n%s' % lib.pformat(self.interface_stats( )))
      log.info(":".join([ log_head, log_detail, "BEGIN ", "first poll" ]))
      size = self.poll_size( )
      end_poll = ':'.join( [ log_head, log_detail,
                             "END first poll %s" % (size),
                             "SHOULD DOWNLOAD ", str(size != 0) ] )
      log.info(end_poll)
      if size == 0:
        break
      
      seg_stats = ( len(segments), sum(map(len, segments)), len(raw) )
      log_detail = segs_vs_raw.format(*seg_stats)
      log.info("%s:download the size? %s:%s" % (log_head, size, log_detail))
          
      while size > 14:
        seg_stats = ( len(segments), sum(map(len, segments)), len(raw) )
        log_detail = segs_vs_raw.format(*seg_stats)
        log_head = "XXX:clear_buffer[attempt][%s]" % (attempt)
        log.info( ':'.join([ "%s size:%s" % (log_head, size),
                             log_detail,
                             "clear_buffer BUFFER self.download( )" ]))
        try:
          segment = self.download( )
          raw.extend(segment)
          segments.append(segment)
          seg_stats = ( len(segments), sum(map(len, segments)), len(raw) )
          log_detail = segs_vs_raw.format(*seg_stats)
          log.info(":".join([ "%s:tx:found" % (log_head),
                              log_detail,
                              'len(raw)', str(len(raw)),
                              'expected', str(size),
                              'len(segment)', str(len(segment)) ]))
        except BadCRC, e:
          seg_stats = ( len(segments), sum(map(len, segments)), len(raw) )
          log_detail = segs_vs_raw.format(*seg_stats)
          log.critical('%s:IGNORING:%s:%s' % (log_head, log_detail, e))

        seg_stats = ( len(segments), sum(map(len, segments)), len(raw) )
        log_detail = segs_vs_raw.format(*seg_stats)
        log.info(':'.join([ "%s downloaded %s segment" % (log_head, len(raw)),
                            log_detail,
                            "RAW:\n%s" % lib.hexdump(raw) ]))
        size = self.poll_size( )

      log.debug("INTERFACE STATS:\n%s" % lib.pformat(self.interface_stats( )))
      if raw:
        return raw
    if size == 0:
      log.info("\n".join([ "%s:END:no data:INTERFACE STATS" % (log_head),
                           lib.pformat(self.interface_stats( )) ]))

  def transmit_packet(self, command):
    """
    Address a pump with a request.
    """
    packet = TransmitPacket(command)
    self.command = packet
    self.transmit = packet
    log.info('transmit_packet:write:%r' % (self.command))
    result = self.process( )
    return result

  def open(self):
    """
    Open and get signal strength so everything is ready to go.
    """
    self.link.baudrate = 9600
    self.timer = lib.Timer( )
    for attempt in xrange( 1 ):
      try:
        msg = ':'.join(['PROCESS', 'OPEN', str(self.timer.millis( ))] )
        log.info(msg)
        log.info('%s' % self.product_info( ))
        log.info('%s' % self.product_info( ))
        log.info('get signal strength of %s' % self)
        signal = 0
        while signal < 50:
          signal = self.signal_strength( )
        log.info('we seem to have found a nice signal strength of: %s' % signal)
        return True
      except AckError, e:
        log.info('failed:(%s):\n%s' % (attempt, e))
        raise
    
  def close (self):
    self.link.close( )

  @staticmethod
  def decode_hex (msg, Candidate):
    candidate = Candidate( )
    raw = lib.hexbytes(msg)
    ack, resp = candidate.respond(raw)
    result = candidate.parse(resp)
    return result

if __name__ == '__main__':
  import doctest
  doctest.testmod( )

  import sys
  port = None
  port = sys.argv[1:] and sys.argv[1] or False
  if not port:
    print "usage:\n%s /dev/ttyUSB0" % sys.argv[0]
    sys.exit(1)
  import link
  from pprint import pformat
  logging.basicConfig(stream=sys.stdout, level=logging.DEBUG)
  log.info("howdy! I'm going to take a look at your carelink usb stick.")
  stick = Stick(link.Link(port))
  stick.open( )
  log.info('test fetching product info %s' % stick)
  log.info(pformat(stick.product_info( )))
  log.info('get signal strength of %s' % stick)
  signal = 0
  while signal < 50:
    signal = stick.signal_strength( )
  log.info('we seem to have found a nice signal strength of: %s' % signal)
  log.info("""
    at this point, we could issue remote commands to a medical
    device, let's inspect the interfaces""".strip( ))
  #log.info(pformat(stick.usb_stats( )))
  #log.info(pformat(stick.radio_stats( )))
  log.info(pformat(stick.interface_stats( )))
  """
  size = stick.poll_size( )
  log.info("can we poll the size? %s" % (size))
  if size > 14:
    log.info("DOWNLOADING %s TO CLEAR BUFFER" % size)
    log.info('\n'.join(["can we download ?", lib.hexdump(stick.download( ))]))
  """
  log.info("CLEAR BUFFERS")
  extra = stick.clear_buffer( )
  if extra:
    log.info(lib.hexdump(extra))
  else:
    log.info("NO PENDING BUFFER")
  log.info("DONE CLEARING BUFFERS")
  log.info("INTERFACE STATS:\n%s" % pformat(stick.interface_stats( )))
  log.info("howdy! all done looking at the stick")

#####
# EOF




import sys
from subprocess import Popen, PIPE

def in_use (device):
  if 'windows' in sys.platform:
    # TODO: use Handle
    # http://stackoverflow.com/questions/18059798/windows-batch-equivalent-of-fuser-k-folder
    # https://technet.microsoft.com/en-us/sysinternals/bb896655
    return False
  pipe = Popen(['fuser', device], stdout=PIPE, stderr=PIPE)
  stdout, stderr = pipe.communicate( )
  return stdout is not ""

if __name__ == '__main__':
  from scan import scan
  candidate = (sys.argv[1:2] or [scan( )]).pop( )
  print in_use(candidate)

